# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for Numpy-based tensors."""

from __future__ import division

import operator
import math
import pytest

import odl
from odl.core.set.space import LinearSpaceTypeError
from odl.core.space.entry_points import TENSOR_SPACE_IMPLS
from odl.core.util.testutils import (
    all_almost_equal, all_equal, noise_array, noise_element, noise_elements,
    isclose, simple_fixture)
from odl.core.array_API_support import lookup_array_backend
from odl.core.util.pytest_config import IMPL_DEVICE_PAIRS

from odl.core.util.dtype_utils import is_complex_dtype

# --- Test helpers --- #

# Functions to return arrays and classes corresponding to impls. Extend
# when a new impl is available.


def _pos_array(space):
    """Create an array with positive real entries in ``space``."""
    ns = space.array_backend.array_namespace
    return ns.abs(noise_array(space)) + 0.1

# --- Pytest fixtures --- #

exponent = simple_fixture('exponent', [2.0, 1.0, float('inf'), 0.5, 1.5])

setitem_indices_params = [
    0, [1], (1,), (0, 1), (0, 1, 2), slice(None), slice(None, None, 2),
    (0, slice(None)), (slice(None), 0, slice(None, None, 2))]
setitem_indices = simple_fixture('indices', setitem_indices_params)

getitem_indices_params = (setitem_indices_params +
                          [([0, 1, 1, 0], [0, 1, 1, 2]), (Ellipsis, None)])
getitem_indices = simple_fixture('indices', getitem_indices_params)

DEFAULT_SHAPE = (3,4)

@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def tspace(request, odl_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_floating_dtype, 
        impl=impl, 
        device=device
        )

@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def floating_tspace(request, odl_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_floating_dtype, 
        impl=impl, 
        device=device
        )

@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def real_tspace(request, odl_real_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_real_floating_dtype, 
        impl=impl, 
        device=device
        )

@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def scalar_tspace(request, odl_scalar_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_scalar_dtype, 
        impl=impl, 
        device=device
        )

# --- Tests --- #
def test_init_tspace(floating_tspace):
    shape  = floating_tspace.shape
    impl   = floating_tspace.impl
    dtype  = floating_tspace.dtype
    device = floating_tspace.device

    # Weights
    constant_weighting = odl.core.space_weighting(
        impl,
        weight = 1.5
        )
    array_weighting    = odl.core.space_weighting(
        impl,
        device, 
        weight = _pos_array(odl.rn(
            shape, 
            impl=impl, dtype=dtype, device=device
            )
        ))
    
    tspace_impl = TENSOR_SPACE_IMPLS[impl]

    for weighting in [constant_weighting, array_weighting, None]:
        tspace_impl(
            DEFAULT_SHAPE, 
            dtype=dtype, 
            device=device, 
            weighting=weighting
            )

def test_properties(odl_impl_device_pairs):
    """Test that the space and element properties are as expected."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space(DEFAULT_SHAPE, dtype='float32', exponent=1, weighting=2,impl=impl, device=device)
    x = space.element()

    ns = space.array_namespace
    assert x.space is space
    assert x.ndim == space.ndim == 2
    assert x.dtype == space.dtype == getattr(ns, 'float32')
    assert x.size == space.size == 12
    assert x.shape == space.shape == DEFAULT_SHAPE
    assert x.itemsize == 4
    assert x.nbytes == 4 * 3 * 4
    assert x.device == device


def test_size(odl_tspace_impl, odl_scalar_dtype):
    """Test that size handles corner cases appropriately."""
    impl = odl_tspace_impl
    space = odl.tensor_space(DEFAULT_SHAPE,dtype=odl_scalar_dtype,  impl=impl)
    assert space.size == 12
    assert type(space.size) == int

    # Size 0
    space = odl.tensor_space((), impl=impl)
    assert space.size == 0
    assert type(space.size) == int

    # Overflow test
    large_space = odl.tensor_space((10000,) * 3, impl=impl)
    assert large_space.size == 10000 ** 3
    assert type(space.size) == int

def test_equals_space(tspace):
    """Test equality check of spaces."""
    impl = tspace.impl
    device = tspace.device
    dtype=tspace.dtype
    space = odl.tensor_space(3, impl=impl, dtype=dtype, device=device)
    same_space = odl.tensor_space(3, impl=impl, dtype=dtype, device=device)
    other_space = odl.tensor_space(4, impl=impl, dtype=dtype, device=device)

    assert space == space
    assert space == same_space
    assert space != other_space
    assert hash(space) == hash(same_space)
    assert hash(space) != hash(other_space)


def test_equals_elem(odl_impl_device_pairs):
    """Test equality check of space elements."""
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, exponent=2, impl=impl, device=device)
    r3_1 = odl.rn(3, exponent=1, impl=impl, device=device)
    r4 = odl.rn(4, exponent=2, impl=impl, device=device)
    r3_elem = r3.element([1, 2, 3])
    r3_same_elem = r3.element([1, 2, 3])
    r3_other_elem = r3.element([2, 2, 3])
    r3_1_elem = r3_1.element([1, 2, 3])
    r4_elem = r4.element([1, 2, 3, 4])

    assert r3_elem == r3_elem
    assert r3_elem == r3_same_elem
    assert r3_elem != r3_other_elem
    assert r3_elem != r3_1_elem
    assert r3_elem != r4_elem


def test_tspace_astype(odl_impl_device_pairs):
    """Test creation of a space counterpart with new dtype."""
    impl, device = odl_impl_device_pairs
    real_space = odl.rn(DEFAULT_SHAPE, impl=impl, device=device)
    int_space = odl.tensor_space(DEFAULT_SHAPE, dtype=int, impl=impl, device=device)
    assert real_space.astype(int) == int_space

    # Test propagation of weightings and the `[real/complex]_space` properties
    real = odl.rn(DEFAULT_SHAPE, weighting=1.5, impl=impl, device=device)
    cplx = odl.cn(DEFAULT_SHAPE, weighting=1.5, impl=impl, device=device)
    real_s = odl.rn(DEFAULT_SHAPE, weighting=1.5, dtype='float32', impl=impl, device=device)
    cplx_s = odl.cn(DEFAULT_SHAPE, weighting=1.5, dtype='complex64', impl=impl, device=device)

    # Real
    assert real.astype('float32') == real_s
    assert real.astype('float64') is real
    assert real.real_space is real
    assert real.astype('complex64') == cplx_s
    assert real.astype('complex128') == cplx
    assert real.complex_space == cplx

    # Complex
    assert cplx.astype('complex64') == cplx_s
    assert cplx.astype('complex128') is cplx
    assert cplx.real_space == real
    assert cplx.astype('float32') == real_s
    assert cplx.astype('float64') == real
    assert cplx.complex_space is cplx


def _test_lincomb(space, a, b, discontig):
    """Validate lincomb against direct result using arrays."""
    # Set slice for discontiguous arrays and get result space of slicing
    # What the actual fuck
    if discontig:
        slc = tuple(
            [slice(None)] * (space.ndim - 1) + [slice(None, None, 2)]
        )
        res_space = space.element()[slc].space
    else:
        res_space = space

    # Unaliased arguments
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]
    zarr[:] = a * xarr + b * yarr
    res_space.lincomb(a, x, b, y, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # First argument aliased with output
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * zarr + b * yarr
    res_space.lincomb(a, z, b, y, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # Second argument aliased with output
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * xarr + b * zarr
    res_space.lincomb(a, x, b, z, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # Both arguments aliased with each other
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * xarr + b * xarr
    res_space.lincomb(a, x, b, x, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # All aliased
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * zarr + b * zarr
    res_space.lincomb(a, z, b, z, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])


def test_lincomb(tspace):
    """Validate lincomb against direct result using arrays and some scalars."""
    scalar_values = [0, 1, -1, 3.41]
    for a in scalar_values:
        for b in scalar_values:
            _test_lincomb(tspace, a, b, discontig=False)


def test_lincomb_discontig(odl_impl_device_pairs):
    """Test lincomb with discontiguous input."""
    impl, device = odl_impl_device_pairs

    scalar_values = [0, 1, -1, 3.41]

    # Use small size for small array case
    tspace = odl.rn(DEFAULT_SHAPE, impl=impl, device=device)

    for a in scalar_values:
        for b in scalar_values:
            _test_lincomb(tspace, a, b, discontig=True)

    # Use medium size to test fallback impls
    tspace = odl.rn((30, 40), impl=impl, device=device)

    for a in scalar_values:
        for b in scalar_values:
            _test_lincomb(tspace, a, b, discontig=True)


def test_lincomb_exceptions(tspace):
    """Test whether lincomb raises correctly for bad output element."""
    other_space = odl.rn((4, 3), impl=tspace.impl)

    other_x = other_space.zero()
    x, y, z = tspace.zero(), tspace.zero(), tspace.zero()

    with pytest.raises(LinearSpaceTypeError):
        tspace.lincomb(1, other_x, 1, y, z)

    with pytest.raises(LinearSpaceTypeError):
        tspace.lincomb(1, y, 1, other_x, z)

    with pytest.raises(LinearSpaceTypeError):
        tspace.lincomb(1, y, 1, z, other_x)

    with pytest.raises(LinearSpaceTypeError):
        tspace.lincomb([], x, 1, y, z)

    with pytest.raises(LinearSpaceTypeError):
        tspace.lincomb(1, x, [], y, z)


def test_multiply(tspace):
    """Test multiply against direct array multiplication."""
    [x_arr, y_arr, out_arr], [x, y, out] = noise_elements(tspace, 3)
    out_arr = x_arr * y_arr

    tspace.multiply(x, y, out)
    assert all_almost_equal([x_arr, y_arr, out_arr], [x, y, out])

    # member method
    [x_arr, y_arr, out_arr], [x, y, out] = noise_elements(tspace, 3)
    out_arr = x_arr * y_arr

    x.multiply(y, out=out)
    assert all_almost_equal([x_arr, y_arr, out_arr], [x, y, out])


def test_multiply_exceptions(tspace):
    """Test if multiply raises correctly for bad input."""
    other_space = odl.rn((4, 3))

    other_x = other_space.zero()
    x, y = tspace.zero(), tspace.zero()

    with pytest.raises(AssertionError):
        tspace.multiply(other_x, x, y)

    with pytest.raises(AssertionError):
        tspace.multiply(x, other_x, y)

    with pytest.raises(AssertionError):
        tspace.multiply(x, y, other_x)


def test_power(tspace):
    """Test ``**`` against direct array exponentiation."""
    [x_arr, y_arr], [x, y] = noise_elements(tspace, n=2)
    y_pos = tspace.element(odl.abs(y) + 0.1)
    ns = tspace.array_namespace
    y_pos_arr = ns.abs(y_arr) + 0.1

    # Testing standard positive integer power out-of-place and in-place
    assert all_almost_equal(x ** 2, x_arr ** 2)
    y **= 2
    y_arr **= 2
    assert all_almost_equal(y, y_arr)
    if tspace.impl == 'pytorch' and is_complex_dtype(tspace.dtype):
        pass
    else:
        # Real number and negative integer power
        assert all_almost_equal(y_pos ** 1.3, y_pos_arr ** 1.3)
        assert all_almost_equal(y_pos ** (-3), y_pos_arr ** (-3))
        y_pos **= 2.5
        y_pos_arr **= 2.5
        assert all_almost_equal(y_pos, y_pos_arr)

        # Array raised to the power of another array, entry-wise
        assert all_almost_equal(y_pos ** x, y_pos_arr ** x_arr)
        y_pos **= x.real
        y_pos_arr **= x_arr.real
        assert all_almost_equal(y_pos, y_pos_arr)


def test_unary_ops(tspace):
    """Verify that the unary operators (`+x` and `-x`) work as expected."""
    for op in [operator.pos, operator.neg]:
        x_arr, x = noise_elements(tspace)

        y_arr = op(x_arr)
        y = op(x)

        assert all_almost_equal([x, y], [x_arr, y_arr])


def test_scalar_operator(tspace, odl_arithmetic_op):
    """Verify binary operations with scalars.

    Verifies that the statement y = op(x, scalar) gives equivalent results
    to NumPy.
    """
    op = odl_arithmetic_op
    if op in (operator.truediv, operator.itruediv):
        ndigits = int(-math.log10(tspace.finfo().resolution) // 2)
    else:
        ndigits = int(-math.log10(tspace.finfo().resolution))

    for scalar in [-31.2, -1, 0, 1, 2.13]:
        x_arr, x = noise_elements(tspace)
        # Left op
        if scalar == 0 and op in [operator.truediv, operator.itruediv]:
            # Check for correct zero division behaviour
            with pytest.raises(ZeroDivisionError):
                y = op(x, scalar)
        else:
            y_arr = op(x_arr, scalar)
            y = op(x, scalar)
            assert all_almost_equal([x, y], [x_arr, y_arr], ndigits)

        # right op
        x_arr, x = noise_elements(tspace)

        y_arr = op(scalar, x_arr)
        y = op(scalar, x)

        
        assert all_almost_equal([x, y], [x_arr, y_arr], ndigits)


def test_binary_operator(tspace, odl_arithmetic_op):
    """Verify binary operations with tensors.

    Verifies that the statement z = op(x, y) gives equivalent results
    to NumPy.
    """
    op = odl_arithmetic_op
    if op in (operator.truediv, operator.itruediv):
        ndigits = int(-math.log10(tspace.finfo().resolution) // 2)
    else:
        ndigits = int(-math.log10(tspace.finfo().resolution))

    [x_arr, y_arr], [x, y] = noise_elements(tspace, 2)

    # non-aliased left
    z_arr = op(x_arr, y_arr)
    z = op(x, y)

    assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr], ndigits)

    # non-aliased right
    z_arr = op(y_arr, x_arr)
    z = op(y, x)

    assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr], ndigits)

    # aliased operation
    z_arr = op(x_arr, x_arr)
    z = op(x, x)

    assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr], ndigits)


def test_assign(tspace):
    """Test the assign method using ``==`` comparison."""
    x = noise_element(tspace)
    x_old = x
    y = noise_element(tspace)

    y.assign(x)

    assert y == x
    assert y is not x
    assert x is x_old

    # test alignment
    x *= 2
    assert y != x


def test_inner(tspace):
    """Test the inner method against numpy.vdot."""
    xarr, xd = noise_elements(tspace)
    yarr, yd = noise_elements(tspace)
    ns = tspace.array_namespace
    # TODO: add weighting
    correct_inner = tspace.array_backend.to_cpu(
        ns.vdot(yarr.ravel(), xarr.ravel())
        )
    assert tspace.inner(xd, yd) == pytest.approx(correct_inner)
    assert xd.inner(yd) == pytest.approx(correct_inner)


def test_inner_exceptions(tspace):
    """Test if inner raises correctly for bad input."""
    other_space = odl.rn((4, 3))
    other_x = other_space.zero()
    x = tspace.zero()

    with pytest.raises(LinearSpaceTypeError):
        tspace.inner(other_x, x)

    with pytest.raises(LinearSpaceTypeError):
        tspace.inner(x, other_x)


def test_norm(tspace):
    """Test the norm method against numpy.linalg.norm."""
    xarr, x = noise_elements(tspace)
    xarr, x = noise_elements(tspace)

    ns = tspace.array_namespace
    correct_norm = tspace.array_backend.to_cpu(
        ns.linalg.norm(xarr.ravel())
        )    

    array_backend = tspace.array_backend
    real_dtype = array_backend.identifier_of_dtype(tspace.real_dtype)
    if real_dtype == "float16":
        tolerance = 5e-3
    elif real_dtype == "float32":
        tolerance = 5e-7
    elif real_dtype == "float64" or real_dtype == float:
        tolerance = 1e-15
    elif real_dtype == "float128":
        tolerance = 1e-19
    else:
        raise TypeError(f"No known tolerance for dtype {tspace.dtype}")
    
    assert tspace.norm(x) == pytest.approx(correct_norm, rel=tolerance)
    assert x.norm() == pytest.approx(correct_norm, rel=tolerance)


    correct_norm = ns.linalg.norm(xarr.ravel())

def test_norm_exceptions(tspace):
    """Test if norm raises correctly for bad input."""
    other_space = odl.rn((4, 3))
    other_x = other_space.zero()

    with pytest.raises(LinearSpaceTypeError):
        tspace.norm(other_x)


def test_pnorm(exponent, odl_impl_device_pairs):
    """Test the norm method with p!=2 against numpy.linalg.norm."""
    impl, device = odl_impl_device_pairs
    space_list = [
        odl.rn(DEFAULT_SHAPE, exponent=exponent,device=device, impl=impl),
        odl.cn(DEFAULT_SHAPE, exponent=exponent,device=device, impl=impl)
    ]
    for tspace in space_list:
        xarr, x = noise_elements(tspace)
        ns = tspace.array_namespace
        correct_norm = tspace.array_backend.to_cpu(ns.linalg.norm(xarr.ravel(), ord=exponent))

        assert tspace.norm(x) == pytest.approx(correct_norm)
        assert x.norm() == pytest.approx(correct_norm)


def test_dist(tspace):
    """Test the dist method against numpy.linalg.norm of the difference."""

    [xarr, yarr], [x, y] = noise_elements(tspace, n=2)

    [xarr, yarr], [x, y] = noise_elements(tspace, n=2)

    ns = tspace.array_namespace
    correct_dist = tspace.array_backend.to_cpu(
         ns.linalg.norm((xarr - yarr).ravel())
        )    

    array_backend = tspace.array_backend
    real_dtype = array_backend.identifier_of_dtype(tspace.real_dtype)

    if real_dtype == "float16":
        tolerance = 5e-3
    elif real_dtype == "float32":
        tolerance = 5e-7
    elif real_dtype == "float64" or real_dtype == float:
        tolerance = 1e-15
    elif real_dtype == "float128":
        tolerance = 1e-19

    else:
        raise TypeError(f"No known tolerance for dtype {tspace.dtype}")
    
    assert tspace.dist(x, y) == pytest.approx(correct_dist, rel=tolerance)
    assert x.dist(y) == pytest.approx(correct_dist, rel=tolerance)



# def test_dist_exceptions(odl_tspace_impl):
#     """Test if dist raises correctly for bad input."""
#     for device in AVAILABLE_DEVICES[odl_tspace_impl]:
#         tspace = odl.tensor_space(DEFAULT_SHAPE, impl=odl_tspace_impl, device=device)
#         other_space = odl.rn((4, 3))
#         other_x = other_space.zero()
#         x = tspace.zero()

#         with pytest.raises(LinearSpaceTypeError):
#             tspace.dist(other_x, x)

#         with pytest.raises(LinearSpaceTypeError):
#             tspace.dist(x, other_x)


def test_pdist(odl_impl_device_pairs, exponent):
    """Test the dist method with p!=2 against numpy.linalg.norm of diff."""
    impl, device = odl_impl_device_pairs
    spaces = [
        odl.rn(DEFAULT_SHAPE, exponent=exponent, impl=impl, device=device),
        odl.cn(DEFAULT_SHAPE, exponent=exponent, impl=impl, device=device)
        ]
    # cls = odl.core.space.entry_points.tensor_space_impl(impl)

    # if complex in cls.available_dtypes:
    #     spaces.append(odl.cn(DEFAULT_SHAPE, exponent=exponent, impl=impl))

    for space in spaces:
        [xarr, yarr], [x, y] = noise_elements(space, n=2)
        ns = space.array_namespace
        correct_dist = space.array_backend.to_cpu(ns.linalg.norm((xarr - yarr).ravel(), ord=exponent))
        assert space.dist(x, y) == pytest.approx(correct_dist)
        assert x.dist(y) == pytest.approx(correct_dist)


def test_element_getitem(odl_impl_device_pairs, getitem_indices):
    """Check if getitem produces correct values, shape and other stuff."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space((2, 3, 4), dtype='float32', exponent=1,
                             weighting=2, impl=impl, device=device)
    x_arr, x = noise_elements(space)

    x_arr_sliced = x_arr[getitem_indices]
    sliced_shape = x_arr_sliced.shape
    x_sliced = x[getitem_indices]

    if x_arr_sliced.ndim == 0:
        try:
            assert x_arr_sliced == x_sliced
        except IndexError:
            assert x_arr_sliced[0] == x_sliced
    else:
        assert x_sliced.shape == sliced_shape
        assert all_equal(x_sliced, x_arr_sliced)

        # Check that the space properties are preserved
        sliced_spc = x_sliced.space
        assert sliced_spc.shape == sliced_shape
        assert sliced_spc.dtype == space.dtype
        assert sliced_spc.exponent == space.exponent
        assert sliced_spc.weighting == space.weighting

        # Check that we have a view that manipulates the original array
        # (or not, depending on indexing style)
        x_arr_sliced[:] = 0
        x_sliced[:] = 0
        assert all_equal(x_arr, x)


def test_element_setitem(setitem_indices, odl_impl_device_pairs):
    """Check if setitem produces the same result as NumPy."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space((2, 3, 4), dtype='float32', exponent=1,
                             weighting=2, impl=impl, device=device)
    x_arr, x = noise_elements(space)



    x_arr_sliced = x_arr[setitem_indices]
    sliced_shape = x_arr_sliced.shape

    ns = space.array_namespace
    # Setting values with scalars
    x_arr[setitem_indices] = 2.3
    x[setitem_indices] = 2.3
    assert all_equal(x, x_arr)

    # Setting values with arrays
    rhs_arr = ns.ones(sliced_shape, device=device)
    x_arr[setitem_indices] = rhs_arr
    x[setitem_indices] = rhs_arr
    assert all_equal(x, x_arr)

    # Using a list of lists
    rhs_list = (-ns.ones(sliced_shape, device=device)).tolist()
    if impl != 'pytorch':
        x_arr[setitem_indices] = rhs_list
        x[setitem_indices] = rhs_list
        assert all_equal(x, x_arr)


def test_element_getitem_bool_array(odl_impl_device_pairs):
    """Check if getitem with boolean array yields the same result as NumPy."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space((2, 3, 4), dtype='float32', exponent=1,
                             weighting=2, impl=impl, device=device)
    bool_space = odl.tensor_space((2, 3, 4), dtype=bool, impl=impl, device=device)
    x_arr, x = noise_elements(space)
    cond_arr, cond = noise_elements(bool_space)

    x_arr_sliced = x_arr[cond_arr]
    x_sliced = x[cond]
    assert all_equal(x_arr_sliced, x_sliced)

    # Check that the space properties are preserved
    sliced_spc = x_sliced.space
    assert sliced_spc.shape == x_arr_sliced.shape
    assert sliced_spc.dtype == space.dtype
    assert sliced_spc.exponent == space.exponent
    assert sliced_spc.weighting == space.weighting


def test_element_setitem_bool_array(odl_impl_device_pairs):
    """Check if setitem produces the same result as NumPy."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space((2, 3, 4), dtype='float32', exponent=1,
                             weighting=2, impl=impl, device=device)
    bool_space = odl.tensor_space((2, 3, 4), dtype=bool, impl=impl, device=device)
    x_arr, x = noise_elements(space)
    cond_arr, cond = noise_elements(bool_space)
    ns = space.array_namespace

    x_arr_sliced = x_arr[cond_arr]
    sliced_shape = x_arr_sliced.shape

    # Setting values with scalars
    x_arr[cond_arr] = 2.3
    x[cond] = 2.3
    assert all_equal(x, x_arr)

    # Setting values with arrays
    rhs_arr = ns.ones(sliced_shape, device=device)
    x_arr[cond_arr] = rhs_arr
    x[cond] = rhs_arr
    assert all_equal(x, x_arr)

    # Using a list of lists
    rhs_list = (-ns.ones(sliced_shape, device=device)).tolist()
    if impl == 'pytorch':
        cond_arr = bool_space.array_backend.array_constructor(cond_arr, device=device)
        rhs_list = bool_space.array_backend.array_constructor(rhs_list, device=device)
    else:
        x_arr[cond_arr] = rhs_list
    x[cond] = rhs_list
    assert all_equal(x, x_arr)


def test_transpose(odl_impl_device_pairs):
    """Test the .T property of tensors against plain inner product."""
    impl, device = odl_impl_device_pairs
    spaces = [
        odl.rn(DEFAULT_SHAPE, impl=impl, device=device),
        odl.cn(DEFAULT_SHAPE, impl=impl, device=device)
        ]
    # cls = odl.core.space.entry_points.tensor_space_impl(impl)
    # if complex in cls.available_dtypes():
    #     spaces.append(odl.cn(DEFAULT_SHAPE, impl=impl))

    for space in spaces:
        x = noise_element(space)
        y = noise_element(space)

        # Assert linear operator
        assert isinstance(x.T, odl.Operator)
        assert x.T.is_linear

        # Check result
        assert x.T(y) == pytest.approx(y.inner(x))
        assert all_equal(x.T.adjoint(1.0), x)

        # x.T.T returns self
        assert x.T.T == x

# TODO: SHOULD that be supported???
def test_multiply_by_scalar(tspace):
    """Verify that mult. with NumPy scalars preserves the element type."""
    x = tspace.zero()

    # Simple scalar multiplication, as often performed in user code.
    # This invokes the __mul__ and __rmul__ methods of the ODL space classes.
    # Strictly speaking this operation loses precision if `tspace.dtype` has
    # fewer than 64 bits (Python decimal literals are double precision), but
    # it would be too cumbersome to force a change in the space's dtype.
    output = x * 1.0
    assert x * 1.0 in tspace
    assert 1.0 * x in tspace


def test_member_copy(odl_impl_device_pairs):
    """Test copy method of elements."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space(DEFAULT_SHAPE, dtype='float32', exponent=1, weighting=2, impl=impl, device = device)
    x = noise_element(space)

    y = x.copy()
    assert x == y
    assert y is not x

    # Check that result is not aliased
    x *= 2
    assert x != y


def test_python_copy(odl_impl_device_pairs):
    """Test compatibility with the Python copy module."""
    import copy
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space(DEFAULT_SHAPE, dtype='float32', exponent=1, weighting=2, impl=impl, device = device)
    x = noise_element(space)

    # Shallow copy
    y = copy.copy(x)
    assert x == y
    assert y is not x

    # Check that result is not aliased
    x *= 2
    assert x != y

    # Deep copy
    z = copy.deepcopy(x)
    assert x == z
    assert z is not x

    # Check that result is not aliased
    x *= 2
    assert x != z

def test_conversion_to_scalar(odl_impl_device_pairs):
    """Test conversion of size-1 vectors/tensors to scalars."""
    impl, device = odl_impl_device_pairs
    space = odl.rn(1, impl=impl, device=device)
    # Size 1 real space
    value = 1.5
    element = space.element(value)

    assert int(element) == int(value)
    assert float(element) == float(value)
    assert complex(element) == complex(value)

    # Size 1 complex space
    value = 1.5 + 0.5j
    element = odl.cn(1).element(value)
    assert complex(element) == complex(value)

    # Size 1 multi-dimensional space
    value = 2.1
    element = odl.rn((1, 1, 1)).element(value)
    assert float(element) == float(value)

    # Too large space
    element = odl.rn(2).one()

    with pytest.raises(AssertionError):
        int(element)
    with pytest.raises(AssertionError):
        float(element)
    with pytest.raises(AssertionError):
        complex(element)

def test_bool_conversion(odl_impl_device_pairs):
    """Verify that the __bool__ function works."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space(2, dtype='float32', impl=impl, device = device)
    x = space.element([0, 1])

    with pytest.raises(ValueError):
        bool(x)
    assert odl.any(x)
    assert any(x)
    assert not odl.all(x)
    assert not all(x)

    space = odl.tensor_space(1, dtype='float32', impl=impl, device = device)
    x = space.one()

    assert odl.any(x)
    assert any(x)
    assert odl.all(x)
    assert all(x)

def test_array_wrap_method(odl_impl_device_pairs):
    """Verify that the __array_wrap__ method for NumPy works."""
    impl, device = odl_impl_device_pairs
    space = odl.tensor_space(DEFAULT_SHAPE, dtype='float32', exponent=1, weighting=2,impl=impl, device=device)
    x_arr, x = noise_elements(space)
    y_arr = space.array_namespace.sin(x_arr)
    y = odl.sin(x)  # Should yield again an ODL tensor

    assert all_equal(y, y_arr)
    assert y in space


def test_conj(tspace):
    """Test complex conjugation of tensors."""
    xarr, x = noise_elements(tspace)

    xconj = x.conj()
    assert all_equal(xconj, xarr.conj())

    y = tspace.element()
    xconj = x.conj(out=y)
    assert xconj is y
    assert all_equal(y, xarr.conj())


# --- Weightings (Numpy) --- #


def test_array_weighting_init(real_tspace):
    """Test initialization of array weightings."""
    exponent = 2
    array_backend = real_tspace.array_backend
    impl = real_tspace.impl
    weight_arr = _pos_array(real_tspace)
    weight_elem = real_tspace.element(weight_arr)

    weighting_arr  = odl.core.space_weighting(impl, device=real_tspace.device, weight=weight_arr, exponent=exponent)
    weighting_elem = odl.core.space_weighting(impl, device=real_tspace.device, 
    weight=weight_elem, exponent=exponent)

    assert isinstance(weighting_arr.weight, array_backend.array_type)
    assert isinstance(weighting_elem.weight, array_backend.array_type)


def test_array_weighting_array_is_valid(odl_impl_device_pairs):
    """Test the is_valid method of array weightings."""
    impl, device = odl_impl_device_pairs
    space = odl.rn(DEFAULT_SHAPE, impl=impl, device=device)
    weight_arr = _pos_array(space)

    assert odl.core.space_weighting(impl, weight=weight_arr, device=device)
    # Invalid
    weight_arr[0] = 0
    with pytest.raises(ValueError):
        odl.core.space_weighting(impl, weight=weight_arr, device=device)


def test_array_weighting_equals(odl_impl_device_pairs):
    """Test the equality check method of array weightings."""
    impl, device = odl_impl_device_pairs
    space = odl.rn(5, impl=impl, device=device)
    weight_arr = _pos_array(space)
    weight_elem = space.element(weight_arr)

    weighting_arr = odl.core.space_weighting(impl, weight=weight_arr, device=device)
    weighting_arr2 = odl.core.space_weighting(impl, weight=weight_arr, device=device)
    weighting_elem = odl.core.space_weighting(impl, weight=weight_elem, device=device)
    weighting_elem_copy = odl.core.space_weighting(impl, weight=weight_elem.copy(), device=device)
    weighting_elem2 = odl.core.space_weighting(impl, weight=weight_elem, device=device)
    weighting_other_arr = odl.core.space_weighting(impl, weight=weight_arr +1 , device=device)
    weighting_other_exp = odl.core.space_weighting(impl, weight=weight_arr +1, exponent=1, device=device)

    assert weighting_arr == weighting_arr2
    assert weighting_arr == weighting_elem
    assert weighting_arr == weighting_elem_copy
    assert weighting_elem == weighting_elem2
    assert weighting_arr != weighting_other_arr
    assert weighting_arr != weighting_other_exp


def test_array_weighting_equiv(odl_impl_device_pairs):
    """Test the equiv method of Numpy array weightings."""
    impl, device = odl_impl_device_pairs
    space = odl.rn(5, impl=impl, device=device)
    weight_arr = _pos_array(space)
    weight_elem = space.element(weight_arr)
    different_arr = weight_arr + 1
    w_arr = odl.core.space_weighting(impl, weight=weight_arr, device=device)
    w_elem = odl.core.space_weighting(impl, weight=weight_elem, device=device)
    w_different_arr = odl.core.space_weighting(impl, weight=different_arr, device=device)

    ns = space.array_namespace

    # Equal -> True
    assert w_arr.equiv(w_arr)
    assert w_arr.equiv(w_elem)
    # Different array -> False
    assert not w_arr.equiv(w_different_arr)

    # Test shortcuts in the implementation
    const_arr = ns.ones(space.shape, device=device) * 1.5
    w_const_arr = odl.core.space_weighting(impl, weight=const_arr, device=device)
    w_const = odl.core.space_weighting(impl, weight=1.5, device=device)
    w_wrong_const = odl.core.space_weighting(impl, weight=1, device=device)
    w_wrong_exp = odl.core.space_weighting(impl, weight=1.5, exponent=1, device=device)

    assert w_const_arr.equiv(w_const)
    assert not w_const_arr.equiv(w_wrong_const)
    assert not w_const_arr.equiv(w_wrong_exp)

    # Bogus input
    assert not w_const_arr.equiv(True)
    assert not w_const_arr.equiv(object)
    assert not w_const_arr.equiv(None)


def test_array_weighting_inner(tspace):
    """Test inner product in a weighted space."""
    [xarr, yarr], [x, y] = noise_elements(tspace, 2)

    weight_arr = _pos_array(tspace)
    weighting = odl.core.space_weighting(
        impl = tspace.impl, 
        weight = weight_arr,
        device = tspace.device
        )
    
    ns = tspace.array_namespace

    true_inner = ns.vdot(yarr.ravel(), (xarr * weight_arr).ravel())
    assert weighting.inner(x.data, y.data) == pytest.approx(tspace.array_backend.to_cpu(true_inner))

    # Exponent != 2 -> no inner product, should raise
    with pytest.raises(NotImplementedError):
        odl.core.space_weighting(impl = tspace.impl, weight =weight_arr, exponent=1.0, device = tspace.device).inner(x.data, y.data)


def test_array_weighting_norm(tspace, exponent):
    """Test norm in a weighted space."""
    ns = tspace.array_namespace
    rtol = math.sqrt(ns.finfo(tspace.dtype).resolution)
    xarr, x = noise_elements(tspace)

    weight_arr = _pos_array(tspace)
    weighting = odl.core.space_weighting(impl = tspace.impl, weight=weight_arr, exponent=exponent, device =tspace.device)

    if exponent == float('inf'):
        true_norm = ns.linalg.vector_norm(
            weight_arr * xarr,
            ord=exponent)
    else:
        true_norm = ns.linalg.norm(
            (weight_arr ** (1 / exponent) * xarr).ravel(),
            ord=exponent)

    assert weighting.norm(x.data) == pytest.approx(
        tspace.array_backend.to_cpu(true_norm), rel=rtol)


def test_array_weighting_dist(tspace, exponent):
    """Test dist product in a weighted space."""
    ns = tspace.array_namespace
    rtol = math.sqrt(ns.finfo(tspace.dtype).resolution)
    [xarr, yarr], [x, y] = noise_elements(tspace, n=2)

    weight_arr = _pos_array(tspace)
    weighting = odl.core.space_weighting(impl = tspace.impl, weight=weight_arr, exponent=exponent, device=tspace.device)

    if exponent == float('inf'):
        true_dist = ns.linalg.norm(
            (weight_arr * (xarr - yarr)).ravel(),
            ord=float('inf'))
    else:
        true_dist = ns.linalg.norm(
            (weight_arr ** (1 / exponent) * (xarr - yarr)).ravel(),
            ord=exponent)

    assert weighting.dist(x.data, y.data) == pytest.approx(
        tspace.array_backend.to_cpu(true_dist), rel=rtol)


def test_const_weighting_init(odl_impl_device_pairs, exponent):
    """Test initialization of constant weightings."""
    impl, device = odl_impl_device_pairs
    # Just test if the code runs
    odl.core.space_weighting(impl=impl, weight=1.5, exponent=exponent, device=device)
    with pytest.raises(ValueError):
        odl.core.space_weighting(impl=impl, weight=0, exponent=exponent, device=device)
    with pytest.raises(ValueError):
        odl.core.space_weighting(impl=impl, weight=-1.5, exponent=exponent, device=device)
    with pytest.raises(ValueError):
        odl.core.space_weighting(impl=impl, weight=float('inf'), exponent=exponent, device=device)


def test_const_weighting_comparison(tspace):
    """Test equality to and equivalence with const weightings."""
    odl_tspace_impl = tspace.impl
    ns = tspace.array_namespace
    constant = 1.5

    w_const = odl.core.space_weighting(impl=odl_tspace_impl, weight=constant)
    w_const2 = odl.core.space_weighting(impl=odl_tspace_impl, weight=constant)
    w_other_const = odl.core.space_weighting(impl=odl_tspace_impl, weight=constant+1)
    w_other_exp = odl.core.space_weighting(impl=odl_tspace_impl, weight=constant, exponent = 1)

    const_arr = constant * ns.ones(DEFAULT_SHAPE)

    w_const_arr = odl.core.space_weighting(impl=odl_tspace_impl, weight=const_arr)
    other_const_arr = (constant + 1) * ns.ones(DEFAULT_SHAPE)
    w_other_const_arr =  odl.core.space_weighting(impl=odl_tspace_impl, weight=other_const_arr)

    assert w_const == w_const
    assert w_const == w_const2
    assert w_const2 == w_const
    # Different but equivalent
    assert w_const.equiv(w_const_arr)
    assert w_const != w_const_arr

    # Not equivalent
    assert not w_const.equiv(w_other_exp)
    assert w_const != w_other_exp
    assert not w_const.equiv(w_other_const)
    assert w_const != w_other_const
    assert not w_const.equiv(w_other_const_arr)
    assert w_const != w_other_const_arr

    # Bogus input
    assert not w_const.equiv(True)
    assert not w_const.equiv(object)
    assert not w_const.equiv(None)


def test_const_weighting_inner(tspace):
    """Test inner product with const weighting."""
    [xarr, yarr], [x, y] = noise_elements(tspace, 2)

    ns = tspace.array_namespace

    constant = 1.5
    true_result_const = constant * ns.vecdot(yarr.ravel(), xarr.ravel())

    w_const = odl.core.space_weighting(impl=tspace.impl, weight=constant)

    assert w_const.inner(x, y) == true_result_const

    # Exponent != 2 -> no inner
    w_const = odl.core.space_weighting(impl=tspace.impl, weight=constant, exponent=1)
    with pytest.raises(NotImplementedError):
        w_const.inner(x, y)


def test_const_weighting_norm(tspace, exponent):
    """Test norm with const weighting."""
    xarr, x = noise_elements(tspace)

    ns = tspace.array_namespace

    constant = 1.5
    if exponent == float('inf'):
        factor = constant
    else:
        factor = constant ** (1 / exponent)

    true_norm = float(factor * ns.linalg.norm(xarr.ravel(), ord=exponent))

    w_const =  odl.core.space_weighting(impl=tspace.impl, weight=constant, exponent=exponent)

    array_backend = tspace.array_backend
    real_dtype = array_backend.identifier_of_dtype(tspace.real_dtype)

    if real_dtype == "float16":
        tolerance = 5e-2
    elif real_dtype == "float32":
        tolerance = 5e-6
    elif real_dtype == "float64" or real_dtype == float:
        tolerance = 1e-15
    elif real_dtype == "float128":
        tolerance = 1e-19
    else:
        raise TypeError(f"No known tolerance for dtype {real_dtype}")
    
    # assert w_const.norm(x) == pytest.approx(true_norm, rel=tolerance)
    assert isclose(w_const.norm(x), true_norm, rtol=tolerance)


def test_const_weighting_dist(tspace, exponent):
    """Test dist with const weighting."""
    [xarr, yarr], [x, y] = noise_elements(tspace, 2)

    ns = tspace.array_namespace

    constant = 1.5
    if exponent == float('inf'):
        factor = constant
    else:
        factor = constant ** (1 / exponent)
    true_dist = float(factor * ns.linalg.norm((xarr - yarr).ravel(), ord=exponent))
    w_const = w_const = odl.core.space_weighting(impl=tspace.impl, weight=constant, exponent=exponent)

    array_backend = tspace.array_backend
    real_dtype = array_backend.identifier_of_dtype(tspace.real_dtype)
    if real_dtype == "float16":
        tolerance = 5e-2
    elif real_dtype == "float32":
        tolerance = 5e-7
    elif real_dtype == "float64" or real_dtype == float:
        tolerance = 1e-15
    elif real_dtype == "float128":
        tolerance = 1e-19
    else:
        raise TypeError(f"No known tolerance for dtype {real_dtype}")

    # assert w_const.dist(x, y) == pytest.approx(true_dist, rel=tolerance)
    assert isclose(w_const.dist(x,y), true_dist, rtol=tolerance)



def test_custom_inner(tspace):
    """Test weighting with a custom inner product."""
    ns = tspace.array_namespace
    rtol = math.sqrt(ns.finfo(tspace.dtype).resolution)

    [xarr, yarr], [x, y] = noise_elements(tspace, 2)

    def inner(x, y):
        return ns.linalg.vecdot(y.ravel(), x.ravel())
    
    def inner_lspacelement(x, y):
        return ns.linalg.vecdot(y.data.ravel(), x.data.ravel())

    def dot(x,y):
        return ns.dot(x,y)
    
    w = odl.core.space_weighting(impl=tspace.impl, inner=inner_lspacelement)
    w_same = odl.core.space_weighting(impl=tspace.impl, inner=inner_lspacelement)
    w_other = odl.core.space_weighting(impl=tspace.impl, inner=dot)

    assert w == w
    assert w == w_same
    assert w != w_other

    true_inner = inner(xarr, yarr)
    assert isclose(w.inner(x, y), true_inner)

    true_norm = float(ns.linalg.norm(xarr.ravel()))
    assert isclose(w.norm(x), true_norm)

    true_dist = float(ns.linalg.norm((xarr - yarr).ravel()))
    assert isclose( w.dist(x, y), true_dist, rtol=rtol)

    with pytest.raises(ValueError):
        odl.core.space_weighting(impl=tspace.impl, inner=inner, weight = 1)


def test_custom_norm(tspace):
    """Test weighting with a custom norm."""
    [xarr, yarr], [x, y] = noise_elements(tspace, 2)
    ns = tspace.array_namespace

    def norm(x):
        return ns.linalg.norm(x)

    def other_norm(x):
        return ns.linalg.norm(x, ord=1)

    w = odl.core.space_weighting(impl=tspace.impl, norm=norm)
    w_same = odl.core.space_weighting(impl=tspace.impl, norm=norm)
    w_other = odl.core.space_weighting(impl=tspace.impl, norm=other_norm)

    assert w == w
    assert w == w_same
    assert w != w_other

    with pytest.raises(NotImplementedError):
        w.inner(x, y)

    true_norm = ns.linalg.norm(xarr.ravel())
    pytest.approx(tspace.norm(x), true_norm)

    true_dist = ns.linalg.norm((xarr - yarr).ravel())
    pytest.approx(tspace.dist(x, y), true_dist)

    with pytest.raises(ValueError):
        odl.core.space_weighting(impl=tspace.impl, norm=norm, weight = 1)


def test_custom_dist(tspace):
    """Test weighting with a custom dist."""
    [xarr, yarr], [x, y] = noise_elements(tspace, 2)
    ns = tspace.array_namespace
    def dist(x, y):
        return ns.linalg.norm(x - y)
    
    def dist_lspace_element(x, y):
        return ns.linalg.norm(x.data - y.data)

    def other_dist(x, y):
        return ns.linalg.norm(x - y, ord=1)

    w = odl.core.space_weighting(impl=tspace.impl, dist=dist_lspace_element)
    w_same = odl.core.space_weighting(impl=tspace.impl, dist=dist_lspace_element)
    w_other = odl.core.space_weighting(impl=tspace.impl, dist=other_dist)

    assert w == w
    assert w == w_same
    assert w != w_other

    with pytest.raises(NotImplementedError):
        w.inner(x, y)

    with pytest.raises(NotImplementedError):
        w.norm(x)

    true_dist = ns.linalg.norm((xarr - yarr).ravel())
    pytest.approx(tspace.dist(x, y), true_dist)

    with pytest.raises(ValueError):
        odl.core.space_weighting(impl=tspace.impl, dist=dist, weight = 1)

def test_reduction(tspace):
    """Check that the generated docstrings are not empty."""
    ## In Pytorch 2.6, max and min reductions are not implemented for ComplexDouble dtype
    # <!> Can randomly raise RuntimeWarning: overflow encountered in reduce
    # <!> Can randomly raise AssertionError: assert (nan+8.12708086701316e-308j) == tensor(nan+8.1271e-308j, dtype=torch.complex128)
    x = tspace.element()
    backend = tspace.array_backend.array_namespace
    for name in ['sum', 'prod', 'min', 'max']:
        reduction = getattr(odl, name)
        reduction_arr = getattr(backend, name)
        if name in ['min', 'max'] and is_complex_dtype(tspace.dtype) and tspace.impl == 'pytorch':
            with pytest.raises(RuntimeError):
                assert reduction(x) == reduction_arr(x.data)
        else:
            assert reduction(x) == reduction_arr(x.data)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)
    



# check if environ['SCIPY_ARRAY_API']='1'
import warnings
from os import environ
if 'SCIPY_ARRAY_API' in environ and environ['SCIPY_ARRAY_API']=='1':
        pass
else:
    warnings.warn('The environment variable SCIPY_ARRAY_API must be set to 1. It should be by default when importing odl, but it seems that scipy was imported before odl. If not set, the array API support of scipy will be disabled, meaning that function calls such as ``xlogy`` on GPU will error and throw back pytorch Type errors. Please add the following lines before your first scipy import. \n' \
    'from os import environ \n' \
    'environ["SCIPY_ARRAY_API"]=="1" \n ' \
    '********End of Warning********', stacklevel=2)

import scipy

__all__ = (
    'lambertw',
    'scipy_lambertw',
    'xlogy',
    )

def _helper(operation:str, x1, x2=None, out=None, namespace=scipy.special, **kwargs):
    return x1.space._elementwise_num_operation(
        operation=operation, x1=x1, x2=x2, out=out, namespace=namespace, **kwargs)

def lambertw(x, k=0, tol=1e-8):
    return _helper('lambertw', x, k=k, tol=tol)

def scipy_lambertw(x, k=0, tol=1e-8):
    return scipy.special.lambertw(x, k, tol)

def xlogy(x1, x2, out=None):
    return _helper('xlogy', x1=x1, x2=x2, out=out)


# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Abstract linear vector spaces."""

from __future__ import print_function, division, absolute_import
from builtins import object
from enum import Enum
from dataclasses import dataclass
import numpy as np
from numbers import Number
from typing import Union

from odl.core.set.sets import Field, Set, UniversalSet


__all__ = ('LinearSpace', 'UniversalSpace')


class NumOperationParadigmSupport(Enum):
    NOT_SUPPORTED = 0
    SUPPORTED = 1
    PREFERRED = 2

    @property
    def is_supported(self):
        return self.value > 0

    @property
    def is_preferred(self):
        return self.value > 1

    def __bool__(self):
        return self.is_supported


@dataclass
class SupportedNumOperationParadigms:
    in_place: NumOperationParadigmSupport
    out_of_place: NumOperationParadigmSupport


class LinearSpace(Set):
    """Abstract linear vector space.

    Its elements are represented as instances of the
    `LinearSpaceElement` class.
    """

    def __init__(self, field):
        """Initialize a new instance.

        This method should be called by all inheriting methods so that
        the `field` property of the space is properly set.

        Parameters
        ----------
        field : `Field` or None
            Scalar field of numbers for this space.
        """
        if field is None or isinstance(field, Field):
            self.__field = field
        else:
            raise TypeError('`field` must be a `Field` instance or `None`, '
                            'got {!r}'.format(field))

    @property
    def field(self):
        """Scalar field of numbers for this vector space."""
        return self.__field

    def element(self, inp=None, **kwargs):
        """Create a `LinearSpaceElement` from ``inp`` or from scratch.

        If called without ``inp`` argument, an arbitrary element of the
        space is generated without guarantee of its state.

        If ``inp in self``, this has return ``inp`` or a view of ``inp``,
        otherwise, a copy may or may not occur.

        Parameters
        ----------
        inp : optional
            Input data from which to create the element.
        kwargs :
            Optional further arguments.

        Returns
        -------
        element : `LinearSpaceElement`
            A new element of this space.
        """
        raise NotImplementedError('abstract method')

    @property
    def examples(self):
        """Example elements `zero` and `one` (if available)."""
        # All spaces should yield the zero element
        yield ('Zero', self.zero())

        try:
            yield ('One', self.one())
        except NotImplementedError:
            pass

    @property
    def supported_num_operation_paradigms(self) -> NumOperationParadigmSupport:
        """Specify whether the low-level numerical operations in this space
        support in-place style, whether they support out-of-place style, and
        if one of them is preferred.
        Generally speaking, for fixed-dimensional spaces whose implementation
        is a monolithic array, in-place style is preferrable because it avoids
        allocating new memory.
        By contrast, in spaces that support e.g. adaptive mesh resolution, the
        in-place style may have little advantage because allocation can only
        be decided based on the inputs, and for automatic differentiation it
        may even be necessary to use purely-functional out-of-place style."""
        raise NotImplementedError('abstract method')

    def _lincomb(self, a, x1, b, x2, out):
        """Implement ``out[:] = a * x1 + b * x2``.

        This method is intended to be private. Public callers should
        resort to `lincomb` which is type-checked.
        """
        raise NotImplementedError('abstract method')

    def _dist(self, x1, x2):
        """Return the distance between ``x1`` and ``x2``.

        This method is intended to be private. Public callers should
        resort to `dist` which is type-checked.
        """
        return self.norm(x1 - x2)

    def _norm(self, x):
        """Return the norm of ``x``.

        This method is intended to be private. Public callers should
        resort to `norm` which is type-checked.
        """
        return float(np.sqrt(self.inner(x, x).real))

    def _inner(self, x1, x2):
        """Return the inner product of ``x1`` and ``x2``.

        This method is intended to be private. Public callers should
        resort to `inner` which is type-checked.
        """
        raise LinearSpaceNotImplementedError(
            'inner product not implemented in space {!r}'.format(self))

    def _multiply(self, x1, x2, out):
        """Implement the pointwise multiplication ``out[:] = x1 * x2``.

        This method is intended to be private. Public callers should
        resort to `multiply` which is type-checked.
        """
        raise LinearSpaceNotImplementedError(
            'multiplication not implemented in space {!r}'.format(self))

    def one(self):
        """Return the one (multiplicative unit) element of this space."""
        raise LinearSpaceNotImplementedError(
            '`one` element not implemented in space {!r}'.format(self))

    # Default methods
    def zero(self):
        """Return the zero (additive unit) element of this space."""
        tmp = self.element()
        self.lincomb(0, tmp, 0, tmp, tmp)
        return tmp

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contains : bool
            ``True`` if ``other`` is a `LinearSpaceElement` instance and
            ``other.space`` is equal to this space, ``False`` otherwise.

        Notes
        -----
        This is the strict default where spaces must be equal.
        Subclasses may choose to implement a less strict check.
        """
        return getattr(other, 'space', None) == self

    # Error checking variant of methods
    def lincomb(self, a, x1, b=None, x2=None, out=None):
        """Implement ``out[:] = a * x1 + b * x2``.

        This function implements

            ``out[:] = a * x1``

        or, if ``b`` and ``x2`` are given,

            ``out = a * x1 + b * x2``.

        Parameters
        ----------
        a : `field` element
            Scalar to multiply ``x1`` with.
        x1 : `LinearSpaceElement`
            First space element in the linear combination.
        b : `field` element, optional
            Scalar to multiply ``x2`` with. Required if ``x2`` is
            provided.
        x2 : `LinearSpaceElement`, optional
            Second space element in the linear combination.
        out : `LinearSpaceElement`, optional
            Element to which the result is written.

        Returns
        -------
        out : `LinearSpaceElement`
            Result of the linear combination. If ``out`` was provided,
            the returned object is a reference to it.

        Notes
        -----
        The elements ``out``, ``x1`` and ``x2`` may be aligned, thus a call

            ``space.lincomb(2, x, 3.14, x, out=x)``

        is (mathematically) equivalent to

            ``x = x * (2 + 3.14)``.
        """
        paradigms = self.supported_num_operation_paradigms

        def assert_x2_has_b():
            if b is None and x2 is not None:
                raise ValueError('`x2` provided but not `b`')

        def assert_x1_in_self():
            if x1 not in self:
                raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                           '{!r}'.format(x1, self))

        def assert_x2_in_self():
            if x2 not in self:
                raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                           '{!r}'.format(x2, self))

        def assert_a_in_field():
            if self.field is not None and a not in self.field:
                raise LinearSpaceTypeError('`a` {!r} not an element of the field '
                                           '{!r} of {!r}'
                                           ''.format(a, self.field, self))

        def assert_b_in_field():
            if self.field is not None and b not in self.field:
                raise LinearSpaceTypeError('`b` {!r} not an element of the '
                                           'field {!r} of {!r}'
                                           ''.format(b, self.field, self))

        if out is None:
            if (paradigms.in_place.is_preferred
                 or not paradigms.out_of_place.is_supported):
                out = self.element()
            else:
                assert_a_in_field()
                assert_x1_in_self()
                if b is None:  # Single element
                    assert_x2_has_b()
                    result = self._lincomb(a, x1, 0, x1, out=None)
                    assert(result is not None)
                    return result
                else:  # Two elements
                    assert_b_in_field()
                    assert_x2_in_self()
                    result = self._lincomb(a, x1, b, x2, out=None)
                    assert(result is not None)
                    return result
        elif out not in self:
            raise LinearSpaceTypeError('`out` {!r} is not an element of {!r}'
                                       ''.format(out, self))

        if (not paradigms.in_place.is_supported):
            out.assign(self.lincomb(a, x1, b, x2, out=None), avoid_deep_copy=True)
            return out

        assert_a_in_field()
        assert_x1_in_self()

        if b is None:  # Single element
            assert_x2_has_b()
            self._lincomb(a, x1, 0, x1, out)

        else:  # Two elements
            assert_b_in_field()
            assert_x2_in_self()

            self._lincomb(a, x1, b, x2, out)

        return out

    def dist(self, x1, x2):
        """Return the distance between ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : `LinearSpaceElement`
            Elements whose distance to compute.

        Returns
        -------
        dist : float
            Distance between ``x1`` and ``x2``.
        """
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                       '{!r}'.format(x1, self))
        if x2 not in self:
            raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                       '{!r}'.format(x2, self))
        return float(self._dist(x1, x2))

    def norm(self, x):
        """Return the norm of ``x``.

        Parameters
        ----------
        x : `LinearSpaceElement`
            Element whose norm to compute.

        Returns
        -------
        norm : float
            Norm of ``x``.
        """
        if x not in self:
            raise LinearSpaceTypeError('`x` {!r} is not an element of '
                                       '{!r}'.format(x, self))
        return float(self._norm(x))

    def inner(self, x1, x2):
        """Return the inner product of ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : `LinearSpaceElement`
            Elements whose inner product to compute.

        Returns
        -------
        inner : `LinearSpace.field` element
            Inner product of ``x1`` and ``x2``.
        """
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                       '{!r}'.format(x1, self))
        if x2 not in self:
            raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                       '{!r}'.format(x2, self))
        inner = self._inner(x1, x2)
        if self.field is None:
            return inner
        else:
            return self.field.element(self._inner(x1, x2))

    def _elementwise_num_operation(self, operation:str
                                       , x1: Union["LinearSpaceElement", Number]
                                       , x2: Union[None, "LinearSpaceElement", Number] = None
                                       , out=None
                                       , namespace=None
                                       , **kwargs ):
        """TODO(Justus) rewrite docstring
        Apply the numerical operation implemented by `low_level_method` to
        `x1` and `x2`.
        This is done either in in-place fashion or out-of-place, depending on
        which style is preferred for this space."""

        raise NotImplementedError("abstract method")

    def _element_reduction(self, operation:str
                               , x: "LinearSpaceElement"
                               , **kwargs
                               ):
        raise NotImplementedError("abstract method")

    @property
    def element_type(self):
        """Type of elements of this space (`LinearSpaceElement`)."""
        return LinearSpaceElement

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)
    
    def __pow__(self, shape):
        """Return ``self ** shape``.

        Notes
        -----
        This can be overridden by subclasses in order to give better memory
        coherence or otherwise a better interface.

        Examples
        --------
        Create simple power space:

        >>> r2 = odl.rn(2)
        >>> r2 ** 4
        ProductSpace(rn(2), 4)

        Multiple powers work as expected:

        >>> r2 ** (4, 2)
        ProductSpace(ProductSpace(rn(2), 4), 2)
        """
        from odl.core.space import ProductSpace

        try:
            shape = (int(shape),)
        except TypeError:
            shape = tuple(shape)

        pspace = self
        for n in shape:
            pspace = ProductSpace(pspace, n)

        return pspace

    def __mul__(self, other):
        """Return ``self * other``.

        Notes
        -----
        This can be overridden by subclasses in order to give better memory
        coherence or otherwise a better interface.

        Examples
        --------
        Create simple product space:

        >>> r2 = odl.rn(2)
        >>> r3 = odl.rn(3)
        >>> r2 * r3
        ProductSpace(rn(2), rn(3))
        """
        from odl.core.space import ProductSpace

        if not isinstance(other, LinearSpace):
            raise TypeError('Can only multiply with `LinearSpace`, got {!r}'
                            ''.format(other))

        return ProductSpace(self, other)

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


class LinearSpaceElement(object):

    """Abstract class for `LinearSpace` elements.

    Do not use this class directly -- to create an element of a vector
    space, call the space's `LinearSpace.element` method instead.
    """

    def __init__(self, space):
        """Initialize a new instance.

        All deriving classes must call this method to set the `space`
        property.
        """
        self.__space = space

    @property
    def space(self):
        """Space to which this element belongs."""
        return self.__space

    # Convenience functions
    def assign(self, other, avoid_deep_copy: bool = False):
        """Assign the values of ``other`` to ``self``, like `self[:] = other`.

        If ``avoid_deep_copy`` is True, an attempt is made to reuse the
        storage of ``other`` for ``self``.
        This is in general unsafe (later modifications to ``self`` would
        impact also ``other``, vice versa), but faster and useful particularly
        when ``other`` is ephemeral (rvalue, in C++ terminology).

        Spaces with immutable elements (i.e., `in_place=NOT_SUPPORTED` in
        `supported_num_operation_paradigms`) may opt for performing only a
        shallow copy even if ``avoid_deep_copy`` is False.
        On the other hand, if some type conversion is necessary on ``other``
        then this will usually prompt a copy even if ``avoid_deep_copy``
        is True."""
        if other in self.space:
            self._assign(other, avoid_deep_copy=avoid_deep_copy)
        else:
            self._assign(self.space.element(other),
                         avoid_deep_copy=avoid_deep_copy)

    def _assign(self, other, avoid_deep_copy):
        """Low-level implementation of `self = other`. Assign the values of
        ``other``, which is assumed to be in the same space, to ``self``."""
        raise NotImplementedError(
                f"Abstract method, not implemented for {type(self)}")

    def copy(self):
        """Create an identical (deep) copy of self."""
        result = self.space.element()
        result.assign(self)
        return result

    def lincomb(self, a, x1, b=None, x2=None):
        """Implement ``self[:] = a * x1 + b * x2``.

        Parameters
        ----------
        a : element of ``space.field``
            Scalar to multiply ``x1`` with.
        x1 : `LinearSpaceElement`
            First space element in the linear combination.
        b : element of ``space.field``, optional
            Scalar to multiply ``x2`` with. Required if ``x2`` is
            provided.
        x2 : `LinearSpaceElement`, optional
            Second space element in the linear combination.

        See Also
        --------
        LinearSpace.lincomb
        """
        return self.space.lincomb(a, x1, b, x2, out=self)

    def set_zero(self):
        """Set this element to zero.

        See Also
        --------
        LinearSpace.zero
        """
        self.assign(self.space.zero())
        return self

    # Convenience methods
    # def __iadd__(self, other):
    #     """Implement ``self += other``."""
    #     if self.space.field is None:
    #         return NotImplemented
    #     elif other in self.space:
    #         return self.space.lincomb(1, self, 1, other, out=self)
    #     elif isinstance(other, LinearSpaceElement):
    #         # We do not `return NotImplemented` here since we don't want a
    #         # fallback for in-place. Otherwise python attempts
    #         # `self = self + other` which does not modify self.
    #         raise TypeError('cannot add {!r} and {!r} in-place'
    #                         ''.format(self, other))
    #     elif other in self.space.field:
    #         one = getattr(self.space, 'one', None)
    #         if one is None:
    #             raise TypeError('cannot add {!r} and {!r} in-place'
    #                             ''.format(self, other))
    #         else:
    #             # other --> other * space.one()
    #             return self.space.lincomb(1, self, other, one(), out=self)
    #     else:
    #         try:
    #             other = self.space.element(other)
    #         except (TypeError, ValueError):
    #             raise TypeError('cannot add {!r} and {!r} in-place'
    #                             ''.format(self, other))
    #         else:
    #             return self.__iadd__(other)

    def __add__(self, other):
        """Return ``self + other``."""
        return self.space._elementwise_num_operation(
            'add', self, other
        )
    
    def __sub__(self, other):
        """Return ``self - other``."""
        return self.space._elementwise_num_operation(
            'subtract', self, other
        )
    
    def __mul__(self, other):
        """Return ``self * other``."""
        return self.space._elementwise_num_operation(
            'multiply', self, other
        )
    
    def __truediv__(self, other):
        """Implement ``self / other``."""
        if isinstance(other, Number) and other == 0:
            raise ZeroDivisionError
        return self.space._elementwise_num_operation(
                'divide', self, other
            )
    
    def __floordiv__(self, other):        
        """Implement ``self // other``."""
        return self.space._elementwise_num_operation(
            'floor_divide', self, other
        )

    def __mod__(self, other):        
        """Implement ``self % other``."""
        return self.space._elementwise_num_operation(
            'remainder', self, other
        )
    
    def __pow__(self, other):
        """Implement ``self ** other``, element wise"""
        return self.space._elementwise_num_operation(
            'pow', self, other
        )

    def __radd__(self, other):
        """Return ``other + self``."""
        return self.space._elementwise_num_operation(
            'add', other, self
        )

    def __rsub__(self, other):
        """Return ``other - self``."""
        return self.space._elementwise_num_operation(
            'subtract', other, self
        )
 
    def __rmul__(self, other):
        """Return ``other * self``."""
        return self.space._elementwise_num_operation(
            'multiply', other, self
        )
    
    def __rtruediv__(self, other):
        """Implement ``other / self``."""
        return self.space._elementwise_num_operation(
             'divide', other, self
        )
    
    def __rfloordiv__(self, other):
        """Implement ``other // self``."""
        return self.space._elementwise_num_operation(
            'floor_divide', other, self
        )
    
    def __rmod__(self, other):        
        """Implement ``other % self``."""
        return self.space._elementwise_num_operation(
            'remainder', other, self
        )
    
    def __rpow__(self, other):
        """Implement ``other ** self``, element wise"""
        return self.space._elementwise_num_operation(
            'pow', other, self
        )
    
    def __iadd__(self, other):
        """Implement ``self += other``."""
        self.space._elementwise_num_operation(
            'add', self, other, self
        )
        return self
    
    def __isub__(self, other):
        """Implement ``self -= other``."""
        self.space._elementwise_num_operation(
            'subtract', self, other, self
        )
        return self
    
    def __imul__(self, other):
        """Return ``self *= other``."""
        self.space._elementwise_num_operation(
            'multiply', self, other, self
        )
        return self
    
    def __itruediv__(self, other):
        """Implement ``self /= other``."""
        if isinstance(other, Number) and other == 0:
            raise ZeroDivisionError
        self.space._elementwise_num_operation(
                'divide', self, other, self
            )
        return self
    
    def __ifloordiv__(self, other):
        """Implement ``self //= other``."""
        self.space._elementwise_num_operation(
            'floor_divide', self, other, self
        )
        return self
    
    def __imod__(self, other):
        """Implement ``self %= other``."""
        self.space._elementwise_num_operation(
            'remainder', self, other, self
        )
        return self
    
    def __ipow__(self, other):
        """Implement ``self *= other``, element wise"""
        self.space._elementwise_num_operation(
            'pow', self, other, self
        )
        return self
    
    def __neg__(self):
        """Return ``-self``."""
        if self.space.field is None:
            return NotImplemented
        return (-1) * self

    def __pos__(self):
        """Return ``+self``."""
        return self.copy()

    def __lt__(self, other):
        """Implement ``self < other``."""
        return self.space._elementwise_num_operation('less', self, other)
    
    def __le__(self, other):
        """Implement ``self <= other``."""
        return self.space._elementwise_num_operation('less_equal', self, other)
    
    def __gt__(self, other):
        """Implement ``self > other``."""
        return self.space._elementwise_num_operation('greater', self, other)
    
    def __ge__(self, other):
        """Implement ``self >= other``."""
        return self.space._elementwise_num_operation('greater_equal', self, other)
    
    def __eq__(self, other):
        """Return ``self == other``.

        Two elements are equal if their distance is zero.

        Parameters
        ----------
        other : `LinearSpaceElement`
            Element of this space.

        Returns
        -------
        equals : bool
            ``True`` if the elements are equal ``False`` otherwise.

        See Also
        --------
        LinearSpace.dist

        Notes
        -----
        Equality is very sensitive to numerical errors, thus any
        arithmetic operations should be expected to break equality.

        Examples
        --------
        >>> rn = odl.rn(1, norm=np.linalg.norm)
        >>> x = rn.element([0.1])
        >>> x == x
        True
        >>> y = rn.element([0.1])
        >>> x == y
        True
        >>> z = rn.element([0.3])
        >>> x + x + x == z
        False
        """
        if other is self:
            # Optimization for a common case
            return True
        elif (not isinstance(other, LinearSpaceElement) or
              other.space != self.space):
            # Cannot use (if other not in self.space) since this is not
            # reflexive.
            return False
        else:
            return self.space.dist(self, other) == 0

    def __ne__(self, other):
        """Return ``self != other``."""
        return not self.__eq__(other)

    # Disable hash since vectors are mutable
    __hash__ = None

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)

    def __copy__(self):
        """Return a copy of this element.

        See Also
        --------
        LinearSpace.copy
        """
        return self.copy()

    def __deepcopy__(self, memo):
        """Return a deep copy of this element.

        See Also
        --------
        LinearSpace.copy
        """
        return self.copy()

    def norm(self):
        """Return the norm of this element.

        See Also
        --------
        LinearSpace.norm
        """
        return self.space.norm(self)

    def dist(self, other):
        """Return the distance of ``self`` to ``other``.

        See Also
        --------
        LinearSpace.dist
        """
        return self.space.dist(self, other)

    def inner(self, other):
        """Return the inner product of ``self`` and ``other``.

        See Also
        --------
        LinearSpace.inner
        """
        return self.space.inner(self, other)

    def multiply(self, other, out=None):
        """Return ``out = self * other``.

        If ``out`` is provided, the result is written to it.

        See Also
        --------
        LinearSpace.multiply
        """
        return self.space.multiply(self, other, out=out)

    def divide(self, other, out=None):
        """Return ``out = self / other``.

        If ``out`` is provided, the result is written to it.

        See Also
        --------
        LinearSpace.divide
        """
        return self.space.divide(self, other, out=out)

    @property
    def T(self):
        """This element's transpose, i.e. the functional ``<. , self>``.

        Returns
        -------
        transpose : `InnerProductOperator`

        Notes
        -----
        This function is only defined in inner product spaces.

        In a complex space, the conjugate transpose of is taken instead
        of the transpose only.

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> x = rn.element([1, 2, 3])
        >>> y = rn.element([2, 1, 3])
        >>> x.T(y)
        13.0
        """
        from odl.core.operator import InnerProductOperator
        return InnerProductOperator(self.copy())

    def __array__(self):
        raise RuntimeError("""
           You are trying to convert an ODL object to a plain array, possibly via a NumPy operation. This is not supported in ODL-1.0 anymore because it interferes with the more general Array API and easily leads to confusing results.

           Instead, you should either:

           - Use the ODL operation (e.g. `odl.sin(x)`)
           - Unwrap the raw array contained in the ODL object, as `x.data`
           - Explicitly convert to NumPy (or another raw array type) via DLPack
           """)

    # Give an `Element` a higher priority than any NumPy array type. This
    # forces the usage of `__op__` of `Element` if the other operand
    # is a NumPy object (applies also to scalars!).
    __array_priority__ = 1000000.0


class UniversalSpace(LinearSpace):

    """A dummy linear space class.

    Mostly raising `LinearSpaceNotImplementedError`.
    """

    def __init__(self):
        """Initialize a new instance."""
        super(UniversalSpace, self).__init__(field=UniversalSet())

    def element(self, inp=None):
        """Dummy element creation method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _lincomb(self, a, x1, b, x2, out):
        """Dummy linear combination.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _dist(self, x1, x2):
        """Dummy distance method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _norm(self, x):
        """Dummy norm method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _inner(self, x1, x2):
        """Dummy inner product method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _multiply(self, x1, x2, out):
        """Dummy multiplication method.

        raises `LinearSpaceNotImplementedError`."""
        raise LinearSpaceNotImplementedError

    def _divide(self, x1, x2, out):
        """Dummy division method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def __eq__(self, other):
        """Return ``self == other``.

        Dummy check, ``True`` for any `LinearSpace`.
        """
        return isinstance(other, LinearSpace)

    def __contains__(self, other):
        """Return ``other in self``.

        Dummy membership check, ``True`` for any `LinearSpaceElement`.
        """
        return isinstance(other, LinearSpaceElement)


class LinearSpaceTypeError(TypeError):
    """Exception for type errors in `LinearSpace`'s.

    This exception is raised when the wrong type of element is fed to
    `LinearSpace.lincomb` and related functions.
    """


class LinearSpaceNotImplementedError(NotImplementedError):
    """Exception for unimplemented functionality in `LinearSpace`'s.

    This exception is raised when a method is called in `LinearSpace`
    that has not been defined in a specific space.
    """


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for normalization of user input."""

from __future__ import print_function, division, absolute_import
import numpy as np


__all__ = ('normalized_scalar_param_list', 'normalized_index_expression',
           'normalized_nodes_on_bdry', 'normalized_axes_tuple',
           'safe_int_conv')


def normalized_scalar_param_list(param, length, param_conv=None,
                                 keep_none=True, return_nonconv=False):
    """Return a list of given length from a scalar parameter.

    The typical use case is when a single value or a sequence of
    values is accepted as input. This function makes a list from
    a given sequence or a list of identical elements from a single
    value, with cast to a given parameter type if desired.

    To distinguish a parameter sequence from a single parameter, the
    following rules are applied:

    * If ``param`` is not a sequence, it is treated as a single
      parameter (e.g. ``1``).
    * If ``len(param) == length == 1``, then ``param`` is interpreted
      as a single parameter (e.g. ``[1]`` or ``'1'``).
    * If ``len(param) == length != 1``, then ``param`` is interpreted as
      sequence of parameters.
    * Otherwise, ``param`` is interpreted as a single parameter.

    Note that this function is not applicable to parameters which are
    themselves iterable (e.g. ``'abc'`` with ``length=3`` will be
    interpreted as equivalent to ``['a', 'b', 'c']``).

    Parameters
    ----------
    param :
        Input parameter to turn into a list.
    length : nonnegative int
        Desired length of the output list.
    param_conv : callable, optional
        Conversion applied to each list element. ``None`` means no conversion.
    keep_none : bool, optional
        If ``True``, ``None`` is not converted.
    return_nonconv : bool, optional
        If ``True``, return also the list where no conversion has been
        applied.

    Returns
    -------
    plist : list
        Input parameter turned into a list of length ``length``.
    nonconv : list
        The same as ``plist``, but without conversion. This is only
        returned if ``return_nonconv == True``.

    Examples
    --------
    Turn input into a list of given length, possibly by broadcasting.
    By default, no conversion is performed.

    >>> normalized_scalar_param_list((1, 2, 3), 3)
    [1, 2, 3]
    >>> normalized_scalar_param_list((1, None, 3.0), 3)
    [1, None, 3.0]

    Single parameters are broadcast to the given length.

    >>> normalized_scalar_param_list(1, 3)
    [1, 1, 1]
    >>> normalized_scalar_param_list('10', 3)
    ['10', '10', '10']
    >>> normalized_scalar_param_list(None, 3)
    [None, None, None]

    List entries can be explicitly converted using ``param_conv``. If
    ``None`` should be kept, set ``keep_none`` to ``True``:

    >>> normalized_scalar_param_list(1, 3, param_conv=float)
    [1.0, 1.0, 1.0]
    >>> normalized_scalar_param_list('10', 3, param_conv=int)
    [10, 10, 10]
    >>> normalized_scalar_param_list((1, None, 3.0), 3, param_conv=int,
    ...                              keep_none=True)  # default
    [1, None, 3]

    The conversion parameter can be any callable:

    >>> def myconv(x):
    ...     return False if x is None else bool(x)
    >>> normalized_scalar_param_list((0, None, 3.0), 3, param_conv=myconv,
    ...                              keep_none=False)
    [False, False, True]
    """
    length, length_in = int(length), length
    if length < 0:
        raise ValueError('`length` must be nonnegative, got {}'
                         ''.format(length_in))

    param = np.array(param, dtype=object, copy=True, ndmin=1)
    nonconv_list = list(np.broadcast_to(param, (length,)))

    if len(nonconv_list) != length:
        raise ValueError('sequence `param` has length {}, expected {}'
                         ''.format(len(nonconv_list), length))

    if param_conv is None:
        out_list = list(nonconv_list)
    else:
        out_list = []
        for p in nonconv_list:
            if p is None and keep_none:
                out_list.append(p)
            else:
                out_list.append(param_conv(p))

    if return_nonconv:
        return out_list, nonconv_list
    else:
        return out_list


def normalized_index_expression(indices, shape, int_to_slice=False):
    """Enable indexing with almost Numpy-like capabilities.

    Implements the following features:

    - Usage of general slices and sequences of slices
    - Conversion of `Ellipsis` into an adequate number of ``slice(None)``
      objects
    - Fewer indices than axes by filling up with an `Ellipsis`
    - Error checking with respect to a given shape
    - Conversion of integer indices into corresponding slices

    Parameters
    ----------
    indices : int, `slice`, `Ellipsis` or sequence of these
        Index expression to be normalized.
    shape : sequence of ints
        Target shape for error checking of out-of-bounds indices.
        Also needed to determine the number of axes.
    int_to_slice : bool, optional
        If ``True``, turn integers into corresponding slice objects.

    Returns
    -------
    normalized : tuple of ints or `slice`'s
        Normalized index expression

    Examples
    --------
    Sequences are turned into tuples. We can have at most as many entries
    as the length of ``shape``, but fewer are allowed - the remaining
    list places are filled up by ``slice(None)``:

    >>> normalized_index_expression([1, 2, 3], shape=(3, 4, 5))
    (1, 2, 3)
    >>> normalized_index_expression([1, 2], shape=(3, 4, 5))
    (1, 2, slice(None, None, None))
    >>> normalized_index_expression([slice(2), 2], shape=(3, 4, 5))
    (slice(None, 2, None), 2, slice(None, None, None))
    >>> normalized_index_expression([1, Ellipsis], shape=(3, 4, 5))
    (1, slice(None, None, None), slice(None, None, None))

    By default, integer indices are kept. If they should be converted
    to slices, use ``int_to_slice=True``. This can be useful to guarantee
    that the result of slicing with the returned object is of the same
    type as the array into which is sliced and has the same number of
    axes:

    >>> x = np.zeros(shape=(3, 4, 5))
    >>> idx1 = normalized_index_expression([1, 2, 3], shape=(3, 4, 5),
    ...                                   int_to_slice=True)
    >>> idx1
    (slice(1, 2, None), slice(2, 3, None), slice(3, 4, None))
    >>> x[idx1]
    array([[[ 0.]]])
    >>> idx2 = normalized_index_expression([1, 2, 3], shape=(3, 4, 5),
    ...                                   int_to_slice=False)
    >>> idx2
    (1, 2, 3)
    >>> x[idx2]
    0.0
    """
    ndim = len(shape)
    # Support indexing with fewer indices as indexing along the first
    # corresponding axes. In the other cases, normalize the input.
    if np.isscalar(indices):
        indices = [indices, Ellipsis]
    elif (isinstance(indices, slice) or indices is Ellipsis):
        indices = [indices]

    indices = list(indices)
    if len(indices) < ndim and Ellipsis not in indices:
        indices.append(Ellipsis)

    # Turn Ellipsis into the correct number of slice(None)
    if Ellipsis in indices:
        if indices.count(Ellipsis) > 1:
            raise ValueError('cannot use more than one Ellipsis.')

        eidx = indices.index(Ellipsis)
        extra_dims = ndim - len(indices) + 1
        indices = (indices[:eidx] + [slice(None)] * extra_dims +
                   indices[eidx + 1:])

    # Turn single indices into length-1 slices if desired
    for (i, idx), n in zip(enumerate(indices), shape):
        if np.isscalar(idx):
            if idx < 0:
                idx += n

            if idx >= n:
                raise IndexError('Index {} is out of bounds for axis '
                                 '{} with size {}.'
                                 ''.format(idx, i, n))
            if int_to_slice:
                indices[i] = slice(idx, idx + 1)

    # Catch most common errors
    if any(s.start == s.stop and s.start is not None or
           s.start == n
           for s, n in zip(indices, shape) if isinstance(s, slice)):
        raise ValueError('Slices with empty axes not allowed.')
    if None in indices:
        raise ValueError('creating new axes is not supported.')
    if len(indices) > ndim:
        raise IndexError('too may indices: {} > {}.'
                         ''.format(len(indices), ndim))

    return tuple(indices)


def normalized_nodes_on_bdry(nodes_on_bdry, length):
    """Return a list of 2-tuples of bool from the input parameter.

    This function is intended to normalize a ``nodes_on_bdry`` parameter
    that can be given as a single boolean (global) or as a sequence
    (per axis). Each entry of the sequence can either be a single
    boolean (global for the axis) or a boolean sequence of length 2.

    Parameters
    ----------
    nodes_on_bdry : bool or sequence
        Input parameter to be normalized according to the above scheme.
    length : positive int
        Desired length of the returned list.

    Returns
    -------
    normalized : list of 2-tuples of bool
        Normalized list with ``length`` entries, each of which is a
        2-tuple of boolean values.

    Examples
    --------
    Global for all axes:

    >>> normalized_nodes_on_bdry(True, length=2)
    [(True, True), (True, True)]

    Global per axis:

    >>> normalized_nodes_on_bdry([True, False], length=2)
    [(True, True), (False, False)]

    Mixing global and explicit per axis:

    >>> normalized_nodes_on_bdry([[True, False], False, True], length=3)
    [(True, False), (False, False), (True, True)]
    """
    if isinstance(nodes_on_bdry, bool):
        return [(nodes_on_bdry, nodes_on_bdry)] * length
    elif (length == 1 and len(nodes_on_bdry) == 2
          and all(isinstance(d, bool) for d in nodes_on_bdry)):
        return [nodes_on_bdry[0], nodes_on_bdry[1]]
    elif len(nodes_on_bdry) == length:
        out_list = []

        for i, on_bdry in enumerate(nodes_on_bdry):
            shape_i = np.shape(on_bdry)
            if shape_i == ():
                out_list.append((bool(on_bdry), bool(on_bdry)))
            elif shape_i == (2,):
                out_list.append((bool(on_bdry[0]), bool(on_bdry[1])))
            else:
                raise ValueError('in axis {}: `nodes_on_bdry` has shape {}, '
                                 'expected (2,)'
                                 .format(i, shape_i))
    else:
        raise ValueError('`nodes_on_bdry` has shape {}, expected ({},)'
                         ''.format(shape, length))

    return out_list


def normalized_axes_tuple(axes, ndim):
    """Return a tuple of ``axes`` converted to positive integers.

    This function turns negative entries into equivalent positive
    ones according to standard Python indexing "from the right".

    Parameters
    ----------
    axes : int or sequence of ints
        Single integer or integer sequence of arbitrary length.
        Duplicate entries are not allowed. All entries must fulfill
        ``-ndim <= axis <= ndim - 1``.
    ndim : positive int
        Number of available axes determining the valid axis range.

    Returns
    -------
    axes_list : tuple of ints
        The converted tuple of axes.

    Examples
    --------
    Normalizing a sequence of axes:

    >>> normalized_axes_tuple([0, -1, 2], ndim=3)
    (0, 2, 2)

    Single integer works, too:

    >>> normalized_axes_tuple(-3, ndim=3)
    (0,)
    """
    try:
        axes, axes_in = (int(axes),), axes
    except TypeError:
        axes, axes_in = tuple(int(axis) for axis in axes), axes
        if any(axis != axis_in for axis, axis_in in zip(axes, axes_in)):
            raise ValueError('`axes` may only contain integers, got {}'
                             ''.format(axes_in))
    else:
        if axes[0] != axes_in:
            raise TypeError('`axes` must be integer or sequence, got {}'
                            ''.format(axes_in))

    if len(set(axes)) != len(axes):
        raise ValueError('`axes` may not contain duplicate entries')

    ndim, ndim_in = int(ndim), ndim
    if ndim <= 0:
        raise ValueError('`ndim` must be positive, got {}'.format(ndim_in))

    axes_arr = np.array(axes)
    axes_arr[axes_arr < 0] += ndim
    if np.any((axes_arr < 0) | (axes_arr >= ndim)):
        raise ValueError('all `axes` entries must satisfy -{0} <= axis < {0}, '
                         'got {1}'.format(ndim, axes_in))

    return tuple(axes_arr)


def safe_int_conv(number):
    """Safely convert a single number to integer."""
    try:
        return int(np.array(number).astype(int, casting='safe'))
    except TypeError:
        raise ValueError('cannot safely convert {} to integer'.format(number))


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# This is an attempt to progressively tidy the 'utility.py' module, which is little more than a heap of unstable/unsupported code waiting to crumble.

# Python imports
from numbers import Number
from functools import lru_cache
# Third-Party import
import array_api_compat as xp
# ODL imports
from odl.core.array_API_support import ArrayBackend, lookup_array_backend
from odl.core.array_API_support.utils import _registered_array_backends
from typing import Optional

__all__ = (
    'is_available_dtype',
    'is_numeric_dtype',
    'is_boolean_dtype',
    'is_int_dtype',
    'is_signed_int_dtype',
    'is_unsigned_int_dtype',
    'is_floating_dtype',
    'is_real_floating_dtype',
    'is_complex_dtype',
    'is_real_dtype',
    'real_dtype',
    'complex_dtype'
)

############################# DATA TYPES #############################
# We store all the data types expected by the python array API as lists, and  the maps for conversion as dicts
BOOLEAN_DTYPES = [
    "bool"
    ]

SIGNED_INTEGER_DTYPES = [
    "int8",
    "int16",
    "int32",
    "int64",
]
UNSIGNED_INTEGER_DTYPES = [
    "uint8",
    "uint16",
    "uint32",
    "uint64"
]

INTEGER_DTYPES = SIGNED_INTEGER_DTYPES + UNSIGNED_INTEGER_DTYPES

FLOAT_DTYPES = [
    "float32",
    "float64"
]

COMPLEX_DTYPES = [
    "complex64",
    "complex128"
]

REAL_DTYPES = INTEGER_DTYPES + FLOAT_DTYPES
SCALAR_DTYPES = REAL_DTYPES + COMPLEX_DTYPES
AVAILABLE_DTYPES = BOOLEAN_DTYPES + REAL_DTYPES + COMPLEX_DTYPES 

"""
See type promotion rules https://data-apis.org/array-api/latest/API_specification/type_promotion.html#type-promotion
"""

TYPE_PROMOTION_REAL_TO_COMPLEX = {
    "int8"  : "complex64",
    "int16" : "complex64",
    "int32" : "complex64",
    "int64" : "complex64",
    "uint8" : "complex64",
    "uint16" : "complex64",
    "uint32"  : "complex128",
    "uint64"  : "complex128",
    "float32" : "complex64",
    "float64" : "complex128"
}

TYPE_PROMOTION_COMPLEX_TO_REAL = {
    "complex64"  : "float32",
    "complex128" : "float64"
}

DTYPE_SHORTHANDS = {
    bool: 'bool',
    int: 'int32',
    float: 'float64',
    complex: 'complex128'
}

# These dicts should not be exposed to the users/developpers outside of the module. We rather provide functions that rely on the available array_backends present
def _universal_dtype_identifier(dtype: "str | Number |xp.dtype", array_backend_selection: list[ArrayBackend]=None) -> str :
    """
    Internal helper function to convert a dtype to a backend-agnostic string identifying it semantically.
    (E.g. `'int32'` and `'int64'` and `'float64'` are all possible distinct results, but `np.float64` and
    `torch.float64` and `float` all map to the unique identifier `'float64'`.)
    ambiguity
    The dtype can be provided as a string, a python Number or as an xp.dtype.
    Returns: 
    dtype_as_str (str), dtype identifier 
    Note:
    xp is written here for type hinting, it refers to the fact that the dtype can be provided as a np.float32 or as a torchfloat32, for instance.
    What concrete types of dtype are allowed is determined by `array_backend_selection`.
    If that argument is not provided, all registered backends are taken into consideration.
    """
    # Lazy import 
    from odl.core.space.entry_points import TENSOR_SPACE_IMPLS

    original_dtype = dtype
    shorthand_elaboration = ""
    if dtype in DTYPE_SHORTHANDS:
        dtype = DTYPE_SHORTHANDS[dtype]
        shorthand_elaboration = " (shorthand for {dtype})"

    if isinstance(dtype, (str, Number, type)):
        if dtype in AVAILABLE_DTYPES:
            return dtype
        else:
            raise TypeError(f'The provided dtype {original_dtype}{shorthand_elaboration} is not available. Please use a dtype in {AVAILABLE_DTYPES}')
    if array_backend_selection is None:
        array_backends = _registered_array_backends.values()
    else:
        array_backends = array_backend_selection
    for array_backend in array_backends:
        if dtype in array_backend.available_dtypes.values():
            return array_backend.identifier_of_dtype(dtype)
    raise ValueError(f'The provided dtype {dtype} is not a string, a python Number or a backend-specific dtype of {[be.impl for be in array_backends]}. Please provide either of these.')

@lru_cache
def is_available_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is available."""
    try: 
        _universal_dtype_identifier(dtype)
        return True
    except ValueError or AssertionError:
        return False 
    
@lru_cache
def is_numeric_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is a numeric type."""
    return _universal_dtype_identifier(dtype) in SCALAR_DTYPES

@lru_cache
def is_boolean_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is an boolean type."""
    return _universal_dtype_identifier(dtype) in BOOLEAN_DTYPES

@lru_cache
def is_signed_int_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is an integer type."""
    return _universal_dtype_identifier(dtype) in SIGNED_INTEGER_DTYPES

@lru_cache
def is_unsigned_int_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is an integer type."""
    return _universal_dtype_identifier(dtype) in UNSIGNED_INTEGER_DTYPES

@lru_cache
def is_int_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is an integer type."""
    return _universal_dtype_identifier(dtype) in INTEGER_DTYPES

@lru_cache
def is_floating_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is a floating point type."""
    return _universal_dtype_identifier(dtype) in FLOAT_DTYPES + COMPLEX_DTYPES

@lru_cache
def is_real_floating_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is a floating point type."""
    return _universal_dtype_identifier(dtype) in FLOAT_DTYPES

@lru_cache
def is_complex_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is a complex type."""
    return _universal_dtype_identifier(dtype) in COMPLEX_DTYPES

@lru_cache
def is_real_dtype(dtype: "str | Number |xp.dtype") -> bool:
    """Return ``True`` if ``dtype`` is a real (including integer) type."""
    return _universal_dtype_identifier(dtype) in REAL_DTYPES

def real_dtype(dtype: "str | Number |xp.dtype", default=None, backend: Optional[ArrayBackend] =None) -> str:
    """
    Returns the real counterpart of ``dtype`` if it exists
    Parameters
    ----------
    dtype :
        Input dtype
    default :
        Object to be returned if no real counterpart is found for
        ``dtype``, except for ``None``, in which case an error is raised.
    backend :
        If given, the result dtype will be returned in its version
        specific to that backend (e.g. `torch.float32`), otherwise as a plain string.
    """
    dtype = _universal_dtype_identifier(dtype)
    def for_backend(dt):
        if backend is None:
            return dt
        else:
            try:
                return backend.available_dtypes[dt]
            except KeyError:
                raise ValueError(f"Real version of {dtype} not available on {backend}.")
    if dtype in REAL_DTYPES:
        return for_backend(dtype)
    elif dtype in COMPLEX_DTYPES:
        return for_backend(TYPE_PROMOTION_COMPLEX_TO_REAL[dtype])
    else:
        if default is None:
            raise ValueError(
                f"no real counterpart exists for `dtype` {dtype}")
        else:
            return default
        
def complex_dtype(dtype: "str | Number |xp.dtype", default=None, backend: Optional[ArrayBackend] =None) -> str:
    dtype = _universal_dtype_identifier(dtype)
    def for_backend(dt):
        if backend is None:
            return dt
        else:
            try:
                return backend.available_dtypes[dt]
            except KeyError:
                raise ValueError(f"Complex version of {dtype} not available on {backend}.")
    if dtype in COMPLEX_DTYPES:
        return for_backend(dtype)
    elif dtype in REAL_DTYPES:
        return for_backend(TYPE_PROMOTION_REAL_TO_COMPLEX[dtype])
    else:
        if default is None:
            raise ValueError(
                f"no complex counterpart exists for `dtype` {dtype}")
        else:
            return default




# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tomographic datasets from Mayo Clinic.
In addition to the standard ODL requirements, this library also requires:
    - pydicom
    - Samples from the Mayo dataset, see
    https://ctcicblog.mayo.edu/hubcap/patient-ct-projection-data-library/
"""

from __future__ import division
import numpy as np
import os
import sys
import pydicom
import odl
#import tqdm
from functools import partial

from pydicom.datadict import DicomDictionary, keyword_dict
from odl.core.discr.discr_utils import linear_interpolator
from odl.contrib.datasets.ct.mayo_dicom_dict import new_dict_items

# Update the DICOM dictionary with the extra Mayo tags
DicomDictionary.update(new_dict_items)
# Update the reverse mapping from name to tag
new_names_dict = dict([(val[4], tag) for tag, val in new_dict_items.items()])
keyword_dict.update(new_names_dict)


__all__ = ('load_projections', 'load_reconstruction')


def _read_projections(folder, indices):
    """Read mayo projections from a folder."""
    datasets = []

    # Get the relevant file names
    file_names = sorted([f for f in os.listdir(folder) if f.endswith(".dcm")])

    if len(file_names) == 0:
        raise ValueError('No DICOM files found in {}'.format(folder))

    file_names = file_names[indices]

    data_array = []

    for i, file_name in enumerate(file_names): #enumerate(tqdm.tqdm(file_names,'Loading projection data')):
        # read the file
        try:
            dataset = pydicom.read_file(os.path.join(folder, file_name))
        except:
            print("corrupted file: {}".format(file_name), file=sys.stderr)
            print("error:\n{}".format(sys.exc_info()[1]), file=sys.stderr)
            raise

        if not data_array:
            # Get some required data
            rows = dataset.NumberofDetectorRows
            cols = dataset.NumberofDetectorColumns
            #rescale_intercept = dataset.RescaleIntercept
            #rescale_slope = dataset.RescaleSlope

        else:
            # Sanity checks
            assert rows == dataset.NumberofDetectorRows
            assert cols == dataset.NumberofDetectorColumns
            #assert rescale_intercept == dataset.RescaleIntercept
            #assert rescale_slope == dataset.RescaleSlope

        rescale_intercept = dataset.RescaleIntercept
        rescale_slope = dataset.RescaleSlope
        
        # Load the array as bytes
        proj_array = np.array(np.frombuffer(dataset.PixelData, 'H'),
                              dtype='float32')
        #proj_array = proj_array.reshape([rows, cols], order='F').T
        proj_array = proj_array.reshape([cols, rows])

        # Rescale array
        proj_array *= rescale_slope
        proj_array += rescale_intercept

        data_array.append(proj_array[:, ::-1])
        datasets.append(dataset)
    data_array = np.stack(data_array)
    
    return datasets, data_array


def load_projections(folder, indices=None, use_ffs=True, flat=False, interpolate=True):
    """Load geometry and data stored in Mayo format from folder.
    Parameters
    ----------
    folder : str
        Path to the folder where the Mayo DICOM files are stored.
    indices : optional
        Indices of the projections to load.
        Accepts advanced indexing such as slice or list of indices.
    use_ffs : bool, optional
        If ``True``, a source shift is applied to compensate the flying focal spot.
        Default: ``True``
    flat : bool, optional
        If ``True``, the data is projected on a flat detector.
        Default: ``Flat``
    Returns
    -------
    geometry : ConeBeamGeometry
        Geometry corresponding to the Mayo projector.
    proj_data : `numpy.ndarray`
        Projection data, given as the line integral of the linear attenuation
        coefficient (g/cm^3). Its unit is thus g/cm^2.
    """
    datasets, data_array = _read_projections(folder, indices)

    # Get the angles
    angles = np.array([d.DetectorFocalCenterAngularPosition for d in datasets])
    # Reverse angular axis and set origin at 6 o'clock
    angles = -np.unwrap(angles) - np.pi 

    # Set minimum and maximum corners
    det_shape = np.array([datasets[0].NumberofDetectorColumns,
                          datasets[0].NumberofDetectorRows])
    det_pixel_size = np.array([datasets[0].DetectorElementTransverseSpacing,
                               datasets[0].DetectorElementAxialSpacing])

    # Correct from center of pixel to corner of pixel
    # det_minp = -(np.array(datasets[0].DetectorCentralElement) - 0.5) * det_pixel_size
    # in the implementation of the curved detector it is assumed 
    # that the detector axis are "attached" to the 0-point of the detector,
    # but in the ray transform it is assumed that 
    # the axis are "attached" to the center of the detector.
    # To avoid problems, we make sure that these two point coincide and
    # shift the detector through the detector shift fuction 
    # and rotate the detector axes accordingly 
    det_minp = -det_shape * det_pixel_size / 2 
    det_maxp = det_minp + det_shape * det_pixel_size

    # Select geometry parameters
    src_radius = datasets[0].DetectorFocalCenterRadialDistance
    det_radius = (datasets[0].ConstantRadialDistance -
                  datasets[0].DetectorFocalCenterRadialDistance)
    curv_radius = src_radius + det_radius
    if flat:
        det_curvature_radius = None
    else:
        det_curvature_radius = (curv_radius, None)

    # For unknown reasons, mayo does not include the tag
    # "TableFeedPerRotation", which is what we want.
    # Instead we manually compute the pitch
    table_dist = (datasets[-1].DetectorFocalCenterAxialPosition -
                  datasets[0].DetectorFocalCenterAxialPosition)
    num_rot = (angles[-1] - angles[0]) / (2 * np.pi)
    pitch = table_dist / num_rot
    
    # Convert offset to odl definitions
    offset_along_axis = (datasets[0].DetectorFocalCenterAxialPosition -
                         angles[0] / (2 * np.pi) * pitch)

    # offsets: detector’s focal center -> focal spot
    offset_angular = np.array([d.SourceAngularPositionShift for d in datasets]) 
    offset_radial = np.array([d.SourceRadialDistanceShift for d in datasets]) 
    offset_axial = np.array([d.SourceAxialPositionShift for d in datasets])

    # angles have inverse convention
    shift_d = np.cos(-offset_angular) * (src_radius + offset_radial) - src_radius
    shift_t = np.sin(-offset_angular) * (src_radius + offset_radial)
    # correcting for non-uniform pitch
    det_offset_axial = np.array([d.DetectorFocalCenterAxialPosition for d in datasets])
    shift_z = np.zeros(len(angles))-(det_offset_axial - angles / (2 * np.pi) * pitch - offset_along_axis)
    shift_r = shift_z - offset_axial
    
    # Create partition for detector
    if not flat:
        det_minp[0] /= curv_radius 
        det_maxp[0] /= curv_radius 
    detector_partition = odl.uniform_partition(det_minp, det_maxp, det_shape)

    # Assemble geometry
    angle_partition = odl.nonuniform_partition(angles)
    
    # Flying focal spot
    src_shift_func = None
    if use_ffs:
        shifts = np.transpose(np.vstack([shift_d, shift_t, shift_r]))
        src_shift_func = partial(
            odl.tomo.flying_focal_spot, apart=angle_partition, shifts=shifts)
    else:
        src_shift_func = None
        
    # Detector shift 
    n_agles = len(angles)
    shift_angle =  -((det_shape[0] / 2 - (datasets[0].DetectorCentralElement[0] - 0.5)) 
                   * det_pixel_size[0] / curv_radius)
    shift_d = curv_radius * (np.cos(shift_angle) - 1)
    shift_t = curv_radius * np.sin(shift_angle)
    shifts = np.transpose(np.vstack([np.ones(n_agles) * shift_d, 
                                     np.ones(n_agles) * shift_t, 
                                     shift_z]))
    det_shift_func = partial(
        odl.tomo.flying_focal_spot, apart=angle_partition, shifts=shifts)
    
    # Detector axes (rotate the first axis (1, 0, 0) by shift_angle)
    det_axes_init = [(np.cos(shift_angle), np.sin(shift_angle), 0),
                     (0, 0, 1)]
    
    geometry = odl.tomo.ConeBeamGeometry(angle_partition,
                                         detector_partition,
                                         src_radius=src_radius,
                                         det_radius=det_radius,
                                         pitch=pitch,
                                         det_curvature_radius=det_curvature_radius,
                                         det_axes_init=det_axes_init, 
                                         offset_along_axis=offset_along_axis,
                                         src_shift_func=src_shift_func,
                                         det_shift_func=det_shift_func)

    # number of photons is inverse to the noise variance
    photon_stat = []
    for d in datasets:
        photon_stat.append(d.PhotonStatistics)
    photon_stat = np.expand_dims(np.array(photon_stat), axis=-1)
    
    if flat and interpolate:
        # project the data on a flat grid
        space = odl.discr.uniform_discr([-1,-1,-1], [1,1,1], (3,3,3))
        grid = odl.tomo.RayTransform(space, geometry).range.grid
        data_array = interpolate_flat_grid(data_array, grid, curv_radius)
        
    return geometry, data_array, photon_stat


def interpolate_flat_grid(data_array, range_grid, radial_dist):
    """Return the linear interpolator of the projection data on a flat detector.
    Parameters
    ----------
    data_array : `numpy.ndarray`
        Projection data on the cylindrical detector that should be interpolated.
    range_grid : RectGrid
        Rectilinear grid on the flat detector.
    radial_dist : float
        The constant radial distance, that is the distance between the detector’s
        focal center and its central element.
    Returns
    -------
    proj_data : `numpy.ndarray`
        Interpolated projection data on the flat rectilinear grid.
    """

    # convert coordinates
    theta, up, vp = range_grid.meshgrid

    u = radial_dist * np.arctan(up / radial_dist)
    v = radial_dist / np.sqrt(radial_dist**2 + up**2) * vp

    # Calculate projection data in rectangular coordinates since we have no
    # backend that supports cylindrical
    interpolator = linear_interpolator(
        data_array, range_grid.coord_vectors
    )
    proj_data_interpolated = interpolator((theta, u, v))

    return proj_data_interpolated


def load_reconstruction(folder, slice_start=0, slice_end=None):
    """Load a volume from folder, also returns the corresponding partition.
    Parameters
    ----------
    folder : str
        Path to the folder where the DICOM files are stored.
    slice_start : int
        Index of the first slice to use. Used for subsampling.
    slice_end : int
        Index of the final slice to use.
    Returns
    -------
    partition : `odl.RectPartition`
        Partition describing the geometric positioning of the voxels.
    data : `numpy.ndarray`
        Volumetric data. Scaled such that data = 1 for water (0 HU).
    Notes
    -----
    DICOM data is highly non trivial. Typically, each slice has been computed
    with a slice tickness (e.g. 3mm) but the slice spacing might be
    different from that.
    Further, the coordinates in DICOM is typically the *middle* of the pixel,
    not the corners as in ODL.
    This function should handle all of these peculiarities and give a volume
    with the correct coordinate system attached.
    """
    file_names = sorted([f for f in os.listdir(folder) if f.endswith(".dcm")])

    if len(file_names) == 0:
        raise ValueError('No DICOM files found in {}'.format(folder))

    volumes = []
    datasets = []

    if slice_end is None:
        slice_end = len(file_names)
    file_names = file_names[slice_start:slice_end]

    for file_name in file_names:#tqdm.tqdm(file_names, 'loading volume data'):
        # read the file
        dataset = pydicom.read_file(os.path.join(folder, file_name))

        # Get parameters
        #pixel_size = np.array(dataset.PixelSpacing)
        #pixel_thickness = float(dataset.SliceThickness)
        rows = dataset.Rows
        cols = dataset.Columns

        # Get data array and convert to correct coordinates
        data_array = np.array(np.frombuffer(dataset.PixelData, 'H'),
                              dtype='float32')
        data_array = data_array.reshape([cols, rows], order='C')
        data_array = np.rot90(data_array, -1)

        # Convert from storage type to densities
        # TODO: Optimize these computations
        hu_values = (dataset.RescaleSlope * data_array +
                     dataset.RescaleIntercept)
        mu_water = 0.0192
        densities = (hu_values + 1000) / 1000 * mu_water

        # Store results
        volumes.append(densities)
        datasets.append(dataset)
        
    # since pixel_size and pixel_thickness are same accros volume, 
    # we take it from the first .dicom file    
    pixel_size = np.array(datasets[0].PixelSpacing)
    pixel_thickness = float(datasets[0].SliceThickness)

    voxel_size = np.array(list(pixel_size) + [pixel_thickness])
    shape = np.array([rows, cols, len(volumes)])

    # Compute geometry parameters
    #print(datasets[0].ReconstructionDiameter, datasets[0].TableHeight, datasets[0].ReconstructionTargetCenterPatient, datasets[0].DataCollectionCenterPatient)
    if datasets[0].get("ReconstructionTargetCenterPatient") is not None:
        mid_pt = np.array(datasets[0].ReconstructionTargetCenterPatient)
        mid_pt[1] += dataset.TableHeight
    else:
        mid_pt = np.zeros(3)
    reconstruction_size = datasets[0].ReconstructionDiameter
    min_pt = mid_pt - reconstruction_size / 2
    max_pt = mid_pt + reconstruction_size / 2

    # axis 1 has reversed convention
    #min_pt[1], max_pt[1] = -max_pt[1], -min_pt[1]

    if len(datasets) > 1:
        slice_distance = np.abs(
            float(datasets[1].ImagePositionPatient[2]) -
            float(datasets[0].ImagePositionPatient[2]))
    else:
        # If we only have one slice, we must approximate the distance.
        slice_distance = pixel_thickness

    if 'Siemens'.upper() in dataset.Manufacturer.upper(): 
        if 'HEAD' in dataset.BodyPartExamined.upper():
            min_pt[2] = np.array(datasets[0].ImagePositionPatient)[2]
            max_pt[2] = np.array(datasets[-1].ImagePositionPatient)[2]
        else:
            min_pt[2] = -np.array(datasets[0].ImagePositionPatient)[2]
            max_pt[2] = -np.array(datasets[-1].ImagePositionPatient)[2]
    else:
        if 'HEAD' in dataset.BodyPartExamined.upper():
            min_pt[2] = np.array(datasets[0].ImagePositionPatient)[2]
            max_pt[2] = np.array(datasets[-1].ImagePositionPatient)[2]
        else:
            min_pt[2] = np.array(datasets[-1].ImagePositionPatient)[2]
            max_pt[2] = np.array(datasets[0].ImagePositionPatient)[2]
            volumes = volumes[::-1]
    # The middle of the minimum/maximum slice can be computed from the
    # DICOM attribute "DataCollectionCenterPatient". Since ODL uses corner
    # points (e.g. edge of volume) we need to add half a voxel thickness to
    # both sides. 
    min_pt[2] -= 0.5 * slice_distance
    max_pt[2] += 0.5 * slice_distance
    

    recon_space = odl.uniform_discr(min_pt, max_pt, shape)

    volume = np.transpose(np.array(volumes), (1, 2, 0))

    return recon_space, volume




if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()
        



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Functions for generating standardized examples in spaces."""

from __future__ import print_function, division, absolute_import
from itertools import product

__all__ = ('samples',)


def samples(*sets):
    """Generate samples from the given sets using their ``examples`` method.

    Parameters
    ----------
    set1, ..., setN : `Set` instance
        Set(s) from which to generate the samples.

    Returns
    -------
    samples : `generator`
        Generator that yields tuples of examples from the sets.

    Examples
    --------
    >>> R, C = odl.RealNumbers(), odl.ComplexNumbers()
    >>> for [name_x, x], [name_y, y] in samples(R, C): pass  # use examples
    """
    if len(sets) == 1:
        for example in sets[0].examples:
            yield example
    else:
        generators = [set_.examples for set_ in sets]
        for examples in product(*generators):
            yield examples


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the alternating dual updates solver."""

from __future__ import division
import odl
from odl.solvers.nonsmooth import adupdates
from odl.solvers.nonsmooth.alternating_dual_updates import adupdates_simple

from odl.core.util.testutils import all_almost_equal


# Places for the accepted error when comparing results
HIGH_ACCURACY = 8
LOW_ACCURACY = 4


# def test_adupdates():
#     """Test if the adupdates solver handles the following problem correctly:

#     ( 1  1/2 1/3 1/4) (x_1)   (a)
#     (1/2 1/3 1/4 1/5) (x_2)   (b)
#     (1/3 1/4 1/5 1/6) (x_3) = (c)
#     (1/4 1/5 1/6 1/7) (x_4)   (d).

#     The matrix is the ill-conditined Hilbert matrix, the inverse of which
#     can be given in closed form. If we set

#     (a)   (25/12)                    (x_1) = (1)
#     (b)   (77/60)                    (x_2) = (1)
#     (c) = (19/20)         then       (x_3) = (1)
#     (d)   (319/420),                 (x_4) = (1).

#     We solve the problem

#     min ||Ax - b||^2 + TV(x) s.t. x >= 0

#     for the matrix A, the r.h.s. b as above and the total variation TV, which
#     is given as the (non-cyclic) sum of the distances of consecutive entries
#     of the solution. The solution of this problem is clearly x = (1, 1, 1, 1),
#     since it satisfies the additional constraint and minimizes both terms of
#     the objective function.
#     """

#     mat1 = [[1, 1 / 2, 1 / 3, 1 / 4],
#             [1 / 2, 1 / 3, 1 / 4, 1 / 5]]
#     mat2 = [[1 / 3, 1 / 4, 1 / 5, 1 / 6],
#             [1 / 4, 1 / 5, 1 / 6, 1 / 7]]

#     # Create the linear operators
#     mat1op = odl.MatrixOperator(mat1)
#     mat2op = odl.MatrixOperator(mat2)
#     domain = mat1op.domain
#     tv1 = odl.MatrixOperator([[1.0, -1.0, 0.0, 0.0]])
#     tv2 = odl.MatrixOperator([[0.0, 0.0, 1.0, -1.0]])
#     tv3 = odl.MatrixOperator([[0.0, 1.0, -1.0, 0.0]])
#     nneg = odl.IdentityOperator(domain)
#     ops = [mat1op, mat2op, odl.BroadcastOperator(tv1, tv2), tv3, nneg]

#     # Create inner stepsizes for linear operators
#     mat1s = 1 / mat1op(mat1op.adjoint(mat1op.range.one()))
#     mat2s = 1 / mat2op(mat2op.adjoint(mat2op.range.one()))
#     tv1s = [0.5, 0.5]
#     tv2s = 0.5
#     nnegs = nneg.range.element([1.0, 1.0, 1.0, 1.0])
#     inner_stepsizes = [mat1s, mat2s, tv1s, tv2s, nnegs]

#     expected_solution = domain.element([1, 1, 1, 1])
#     # Create right-hand-sides of the equation
#     rhs1 = mat1op(expected_solution)
#     rhs2 = mat2op(expected_solution)

#     # Create the functionals
#     fid1 = odl.functionals.L2NormSquared(mat1op.range).translated(rhs1)
#     fid2 = odl.functionals.L2NormSquared(mat2op.range).translated(rhs2)
#     reg1 = odl.functionals.L1Norm(tv1.range)
#     reg2 = odl.functionals.L1Norm(tv2.range)
#     reg3 = odl.functionals.L1Norm(tv3.range)
#     ind = odl.functionals.IndicatorNonnegativity(nneg.range)
#     funcs = [fid1, fid2, odl.functionals.SeparableSum(reg1, reg2), reg3, ind]

#     # Start from zero
#     x = tv1.domain.zero()
#     x_simple = tv1.domain.zero()

#     stepsize = 1.0
#     niter = 10

#     adupdates(x, funcs, ops, stepsize, inner_stepsizes, niter)
#     adupdates_simple(x_simple, funcs, ops, stepsize,
#                      inner_stepsizes, niter)
#     assert all_almost_equal(x, x_simple)
#     assert domain.dist(x, expected_solution) < 3e-2


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



from .array_API_support import *
from .discr import *
from .operator import *
from .set import *
from .space import *
from .sparse import *
from .util import *

__all__ = ()

__all__ += array_API_support.__all__
__all__ += util.__all__
__all__ += discr.__all__
__all__ += operator.__all__
__all__ += set.__all__
__all__ += space.__all__



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test iterative solvers."""

from __future__ import division
import odl
from odl.core.util.testutils import all_almost_equal
import pytest
import numpy as np


# Find the valid projectors
@pytest.fixture(scope="module",
                params=['steepest_descent',
                        'adam',
                        'landweber',
                        'conjugate_gradient',
                        'conjugate_gradient_normal',
                        'mlem',
                        'osmlem',
                        'kaczmarz'])
def iterative_solver(request):
    """Return a solver given by a name with interface solve(op, x, rhs)."""
    solver_name = request.param

    if solver_name == 'steepest_descent':
        def solver(op, x, rhs):
            norm2 = op.adjoint(op(x)).norm() / x.norm()
            func = odl.functionals.L2NormSquared(op.domain) * (op - rhs)

            odl.solvers.steepest_descent(func, x, line_search=0.5 / norm2)
    elif solver_name == 'adam':
        def solver(op, x, rhs):
            norm2 = op.adjoint(op(x)).norm() / x.norm()
            func = odl.functionals.L2NormSquared(op.domain) * (op - rhs)

            odl.solvers.adam(func, x, learning_rate=4.0 / norm2, maxiter=150)
    elif solver_name == 'landweber':
        def solver(op, x, rhs):
            norm2 = op.adjoint(op(x)).norm() / x.norm()
            odl.solvers.landweber(op, x, rhs, niter=50, omega=0.5 / norm2)
    elif solver_name == 'conjugate_gradient':
        def solver(op, x, rhs):
            odl.solvers.conjugate_gradient(op, x, rhs, niter=10)
    elif solver_name == 'conjugate_gradient_normal':
        def solver(op, x, rhs):
            odl.solvers.conjugate_gradient_normal(op, x, rhs, niter=10)
    elif solver_name == 'mlem':
        def solver(op, x, rhs):
            odl.solvers.mlem(op, x, rhs, niter=10)
    elif solver_name == 'osmlem':
        def solver(op, x, rhs):
            odl.solvers.osmlem([op, op], x, [rhs, rhs], niter=10)
    elif solver_name == 'kaczmarz':
        def solver(op, x, rhs):
            norm2 = op.adjoint(op(x)).norm() / x.norm()
            odl.solvers.kaczmarz([op, op], x, [rhs, rhs], niter=20,
                                 omega=0.5 / norm2)
    else:
        raise ValueError('solver not valid')

    return solver


# Define some interesting problems
@pytest.fixture(scope="module",
                params=['MatVec',
                        'Identity'])
def optimization_problem(request, odl_impl_device_pairs):
    problem_name = request.param
    impl, device = odl_impl_device_pairs
    if problem_name == 'MatVec':
        # Define problem
        op_arr = np.eye(5) * 5 + np.ones([5, 5])
        space = odl.tensor_space((5,5), impl=impl, device=device)
        op = odl.MatrixOperator(op_arr)

        # Simple right hand side
        rhs = op.range.one()

        # Initial guess
        x = op.domain.element([0.6, 0.8, 1.0, 1.2, 1.4])

        return op, x, rhs
    elif problem_name == 'Identity':
        # Define problem
        space = odl.uniform_discr(0, 1, 5, impl = impl, device=device)
        op = odl.IdentityOperator(space)

        # Simple right hand side
        rhs = op.range.element([0, 0, 1, 0, 0])

        # Initial guess
        x = op.domain.element([0.6, 0.8, 1.0, 1.2, 1.4])

        return op, x, rhs
    else:
        raise ValueError('problem not valid')


def test_solver(optimization_problem, iterative_solver):
    """Test iterative solver for solving some simple problems."""
    op, x, rhs = optimization_problem

    iterative_solver(op, x, rhs)
    assert all_almost_equal(op(x), rhs, ndigits=2)


def test_steepst_descent(odl_impl_device_pairs):
    """Test steepest descent on the rosenbrock function in 3d."""
    impl, device = odl_impl_device_pairs
    space = odl.rn(3, impl = impl, device=device)
    scale = 1  # only mildly ill-behaved
    rosenbrock = odl.functionals.RosenbrockFunctional(space, scale)

    line_search = odl.solvers.BacktrackingLineSearch(
        rosenbrock, 0.1, 0.01)
    x = rosenbrock.domain.zero()
    odl.solvers.steepest_descent(rosenbrock, x, maxiter=40,
                                 line_search=line_search)

    assert all_almost_equal(x, [1, 1, 1], ndigits=2)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Discrete wavelet transformation on L2 spaces."""

from __future__ import absolute_import, division, print_function

import numpy as np

from odl.core.discr import DiscretizedSpace
from odl.core.operator import Operator
from odl.trafos.backends.pywt_bindings import (
    PYWT_AVAILABLE, precompute_raveled_slices, pywt_pad_mode, pywt_wavelet)

__all__ = ('WaveletTransform', 'WaveletTransformInverse')


_SUPPORTED_WAVELET_IMPLS = ()
if PYWT_AVAILABLE:
    _SUPPORTED_WAVELET_IMPLS += ('pywt',)
    import pywt


class WaveletTransformBase(Operator):

    """Base class for discrete wavelet transforms.

    This abstract class is intended to share code between the forward,
    inverse and adjoint wavelet transforms.
    """

    def __init__(self, space, wavelet, nlevels, variant, pad_mode='constant',
                 pad_const=0, impl='pywt', axes=None):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscretizedSpace`
            Domain of the forward wavelet transform (the "image domain").
            In the case of ``variant in ('inverse', 'adjoint')``, this
            space is the range of the operator.
        wavelet : string or `pywt.Wavelet`
            Specification of the wavelet to be used in the transform.
            If a string is given, it is converted to a `pywt.Wavelet`.
            Use `pywt.wavelist` to get a list of available wavelets.

            Possible wavelet families are:

            ``'haar'``: Haar

            ``'db'``: Daubechies

            ``'sym'``: Symlets

            ``'coif'``: Coiflets

            ``'bior'``: Biorthogonal

            ``'rbio'``: Reverse biorthogonal

            ``'dmey'``: Discrete FIR approximation of the Meyer wavelet

        variant : {'forward', 'inverse', 'adjoint'}
            Wavelet transform variant to be created.
        nlevels : positive int, optional
            Number of scaling levels to be used in the decomposition. The
            maximum number of levels can be calculated with
            `pywt.dwtn_max_level`.
            Default: Use maximum number of levels.
        pad_mode : string, optional
            Method to be used to extend the signal.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not repeating the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'pywt_per'``:  like ``'periodic'``-padding but gives the smallest
            possible number of decomposition coefficients.
            Only available with ``impl='pywt'``, See ``pywt.Modes.modes``.

            ``'reflect'``: Reflect at the boundary, without repeating the
            outmost values.

            ``'antisymmetric'``: Anti-symmetric variant of ``symmetric``.

            ``'antireflect'``: Anti-symmetric variant of ``reflect``.

            For reference, the following table compares the naming conventions
            for the modes in ODL vs. PyWavelets::

                ======================= ==================
                          ODL               PyWavelets
                ======================= ==================
                symmetric               symmetric
                reflect                 reflect
                order1                  smooth
                order0                  constant
                constant, pad_const=0   zero
                periodic                periodic
                pywt_per                periodization
                antisymmetric           antisymmetric
                antireflect             antireflect
                ======================= ==================

            See `signal extension modes`_ for an illustration of the modes
            (under the PyWavelets naming conventions).
        pad_const : float, optional
            Constant value to use if ``pad_mode == 'constant'``. Ignored
            otherwise. Constants other than 0 are not supported by the
            ``pywt`` back-end.
        impl : {'pywt'}, optional
            Back-end for the wavelet transform.
        axes : sequence of ints, optional
            Axes over which the DWT that created ``coeffs`` was performed.  The
            default value of ``None`` corresponds to all axes. When not all
            axes are included this is analagous to a batch transform in
            ``len(axes)`` dimensions looped over the non-transformed axes. In
            orther words, filtering and decimation does not occur along any
            axes not in ``axes``.

        References
        ----------
        .. _signal extension modes:
           https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-modes.html
        """
        if not isinstance(space, DiscretizedSpace):
            raise TypeError(
                '`space` {!r} is not a `DiscretizedSpace` instance'
                ''.format(space)
            )

        self.__impl, impl_in = str(impl).lower(), impl
        if self.impl not in _SUPPORTED_WAVELET_IMPLS:
            raise ValueError("`impl` '{}' not supported".format(impl_in))

        if axes is None:
            axes = tuple(range(space.ndim))
        elif np.isscalar(axes):
            axes = (axes,)
        elif len(axes) > space.ndim:
            raise ValueError("too many axes")
        self.axes = tuple(axes)

        if nlevels is None:
            nlevels = pywt.dwtn_max_level(space.shape, wavelet, self.axes)
        self.__nlevels, nlevels_in = int(nlevels), nlevels
        if self.nlevels != nlevels_in:
            raise ValueError('`nlevels` must be integer, got {}'
                             ''.format(nlevels_in))

        self.__impl, impl_in = str(impl).lower(), impl
        if self.impl not in _SUPPORTED_WAVELET_IMPLS:
            raise ValueError("`impl` '{}' not supported".format(impl_in))

        self.__wavelet = getattr(wavelet, 'name', str(wavelet).lower())
        self.__pad_mode = str(pad_mode).lower()
        self.__pad_const = space.field.element(pad_const)

        if self.impl == 'pywt':
            self.pywt_pad_mode = pywt_pad_mode(pad_mode, pad_const)
            self.pywt_wavelet = pywt_wavelet(self.wavelet)
            # determine coefficient shapes (without running wavedecn)
            self._coeff_shapes = pywt.wavedecn_shapes(
                space.shape, wavelet, mode=self.pywt_pad_mode,
                level=self.nlevels, axes=self.axes)
            # precompute slices into the (raveled) coeffs
            self._coeff_slices = precompute_raveled_slices(self._coeff_shapes)
            coeff_size = pywt.wavedecn_size(self._coeff_shapes)
            coeff_space = space.tspace_type(coeff_size, dtype=space.dtype)
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))

        variant, variant_in = str(variant).lower(), variant
        if variant not in ('forward', 'inverse', 'adjoint'):
            raise ValueError("`variant` '{}' not understood"
                             "".format(variant_in))
        self.__variant = variant

        if variant == 'forward':
            super(WaveletTransformBase, self).__init__(
                domain=space, range=coeff_space, linear=True)
        else:
            super(WaveletTransformBase, self).__init__(
                domain=coeff_space, range=space, linear=True)

    @property
    def impl(self):
        """Implementation back-end of this wavelet transform."""
        return self.__impl

    @property
    def nlevels(self):
        """Number of scaling levels in this wavelet transform."""
        return self.__nlevels

    @property
    def wavelet(self):
        """Name of the wavelet used in this wavelet transform."""
        return self.__wavelet

    @property
    def pad_mode(self):
        """Padding mode used for extending input beyond its boundary."""
        return self.__pad_mode

    @property
    def pad_const(self):
        """Value for extension used in ``'constant'`` padding mode."""
        return self.__pad_const

    @property
    def is_orthogonal(self):
        """Whether or not the wavelet basis is orthogonal."""
        return self.pywt_wavelet.orthogonal

    @property
    def is_biorthogonal(self):
        """Whether or not the wavelet basis is bi-orthogonal."""
        return self.pywt_wavelet.biorthogonal

    def scales(self):
        """Get the scales of each coefficient.

        Returns
        -------
        scales : ``range`` element
            The scale of each coefficient, given by an integer. 0 for the
            lowest resolution and self.nlevels for the highest.
        """
        if self.impl == 'pywt':
            if self.__variant == 'forward':
                discr_space = self.domain
                wavelet_space = self.range
            else:
                discr_space = self.range
                wavelet_space = self.domain

            shapes = pywt.wavedecn_shapes(discr_space.shape, self.pywt_wavelet,
                                          mode=self.pywt_pad_mode,
                                          level=self.nlevels, axes=self.axes)
            coeff_list = [np.full(shapes[0], 0)]
            for i in range(1, 1 + len(shapes[1:])):
                coeff_list.append({k: np.full(shapes[i][k], i)
                                   for k in shapes[i].keys()})
            coeffs = pywt.ravel_coeffs(coeff_list, axes=self.axes)[0]
            return wavelet_space.element(coeffs)
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))


class WaveletTransform(WaveletTransformBase):

    """Discrete wavelet transform between discretized Lp spaces."""

    def __init__(self, domain, wavelet, nlevels=None, pad_mode='constant',
                 pad_const=0, impl='pywt', axes=None):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `DiscretizedSpace`
            Domain of the wavelet transform (the "image domain").
        wavelet : string or `pywt.Wavelet`
            Specification of the wavelet to be used in the transform.
            If a string is given, it is converted to a `pywt.Wavelet`.
            Use `pywt.wavelist` to get a list of available wavelets.

            Possible wavelet families are:

            ``'haar'``: Haar

            ``'db'``: Daubechies

            ``'sym'``: Symlets

            ``'coif'``: Coiflets

            ``'bior'``: Biorthogonal

            ``'rbio'``: Reverse biorthogonal

            ``'dmey'``: Discrete FIR approximation of the Meyer wavelet

        nlevels : positive int, optional
            Number of scaling levels to be used in the decomposition. The
            maximum number of levels can be calculated with
            `pywt.dwtn_max_level`.
            Default: Use maximum number of levels.
        pad_mode : string, optional
            Method to be used to extend the signal.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not repeating the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'pywt_per'``:  like ``'periodic'``-padding but gives the smallest
            possible number of decomposition coefficients.
            Only available with ``impl='pywt'``, See ``pywt.Modes.modes``.

            ``'reflect'``: Reflect at the boundary, without repeating the
            outmost values.

            ``'antisymmetric'``: Anti-symmetric variant of ``symmetric``.

            ``'antireflect'``: Anti-symmetric variant of ``reflect``.

            For reference, the following table compares the naming conventions
            for the modes in ODL vs. PyWavelets::

                ======================= ==================
                          ODL               PyWavelets
                ======================= ==================
                symmetric               symmetric
                reflect                 reflect
                order1                  smooth
                order0                  constant
                constant, pad_const=0   zero
                periodic                periodic
                pywt_per                periodization
                antisymmetric           antisymmetric
                antireflect             antireflect
                ======================= ==================

            See `signal extension modes`_ for an illustration of the modes
            (under the PyWavelets naming conventions).
        pad_const : float, optional
            Constant value to use if ``pad_mode == 'constant'``. Ignored
            otherwise. Constants other than 0 are not supported by the
            ``pywt`` back-end.
        impl : {'pywt'}, optional
            Backend for the wavelet transform.
        axes : sequence of ints, optional
            Axes over which the DWT that created ``coeffs`` was performed.  The
            default value of ``None`` corresponds to all axes. When not all
            axes are included this is analagous to a batch transform in
            ``len(axes)`` dimensions looped over the non-transformed axes. In
            orther words, filtering and decimation does not occur along any
            axes not in ``axes``.

        Examples
        --------
        Compute a very simple wavelet transform in a discrete 2D space with
        4 sampling points per axis:

        >>> space = odl.uniform_discr([0, 0], [1, 1], (4, 4))
        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=1, wavelet='haar')
        >>> wavelet_trafo.is_biorthogonal
        True
        >>> data = [[1, 1, 1, 1],
        ...         [0, 0, 0, 0],
        ...         [0, 0, 1, 1],
        ...         [1, 0, 1, 0]]
        >>> decomp = wavelet_trafo(data)
        >>> decomp.shape
        (16,)

        It is also possible to apply the transform only along a subset of the
        axes. Here, we apply a 1D wavelet transfrom along axis 0 for each
        index along axis 1:

        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=1, wavelet='haar', axes=(0,))
        >>> decomp = wavelet_trafo(data)
        >>> decomp.shape
        (16,)

        In general, the size of the coefficients may exceed the size of the
        input data when the wavelet is longer than the Haar wavelet. This
        due to extra coefficients that must be kept for perfect reconstruction.
        No extra boundary coefficients are needed when the edge mode is
        ``"pywt_periodic"`` and the size along each transformed axis is a
        multiple of ``2**nlevels``.

        >>> space = odl.uniform_discr([0, 0], [1, 1], (16, 16))
        >>> space.size
        256
        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=2, wavelet='db2',
        ...     pad_mode='pywt_periodic')
        >>> decomp = wavelet_trafo(np.ones(space.shape))
        >>> decomp.shape
        (256,)
        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=2, wavelet='db2', pad_mode='symmetric')
        >>> decomp = wavelet_trafo(np.ones(space.shape))
        >>> decomp.shape
        (387,)

        References
        ----------
        .. _signal extension modes:
           https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-modes.html
        """
        super(WaveletTransform, self).__init__(
            space=domain, wavelet=wavelet, nlevels=nlevels, variant='forward',
            pad_mode=pad_mode, pad_const=pad_const, impl=impl, axes=axes)

    def _call(self, x):
        """Return wavelet transform of ``x``."""
        if self.impl == 'pywt':
            coeffs = pywt.wavedecn(
                x.data, wavelet=self.pywt_wavelet, level=self.nlevels,
                mode=self.pywt_pad_mode, axes=self.axes)
            return self.range.element(pywt.ravel_coeffs(coeffs, axes=self.axes)[0])
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))

    @property
    def adjoint(self):
        """Adjoint wavelet transform.

        Returns
        -------
        adjoint : `WaveletTransformInverse`
            If the transform is orthogonal, the adjoint is the inverse.

        Raises
        ------
        OpNotImplementedError
            if `is_orthogonal` is ``False``
        """
        if self.is_orthogonal:
            scale = 1 / self.domain.partition.cell_volume
            return scale * self.inverse
        else:
            # TODO: put adjoint here
            return super(WaveletTransform, self).adjoint

    @property
    def inverse(self):
        """Inverse wavelet transform.

        Returns
        -------
        inverse : `WaveletTransformInverse`

        See Also
        --------
        adjoint
        """
        return WaveletTransformInverse(
            range=self.domain, wavelet=self.pywt_wavelet, nlevels=self.nlevels,
            pad_mode=self.pad_mode, pad_const=self.pad_const, impl=self.impl,
            axes=self.axes)


class WaveletTransformInverse(WaveletTransformBase):

    """Discrete inverse wavelet trafo between discrete L2 spaces.

    See Also
    --------
    WaveletTransform
    """

    def __init__(self, range, wavelet, nlevels=None, pad_mode='constant',
                 pad_const=0, impl='pywt', axes=None):
        """Initialize a new instance.

         Parameters
        ----------
        range : `DiscretizedSpace`
            Domain of the forward wavelet transform (the "image domain"),
            which is the range of this inverse transform.
        wavelet : string or `pywt.Wavelet`
            Specification of the wavelet to be used in the transform.
            If a string is given, it is converted to a `pywt.Wavelet`.
            Use `pywt.wavelist` to get a list of available wavelets.

            Possible wavelet families are:

            ``'haar'``: Haar

            ``'db'``: Daubechies

            ``'sym'``: Symlets

            ``'coif'``: Coiflets

            ``'bior'``: Biorthogonal

            ``'rbio'``: Reverse biorthogonal

            ``'dmey'``: Discrete FIR approximation of the Meyer wavelet

        nlevels : positive int, optional
            Number of scaling levels to be used in the decomposition. The
            maximum number of levels can be calculated with
            `pywt.dwtn_max_level`.
            Default: Use maximum number of levels.
        pad_mode : string, optional
            Method to be used to extend the signal.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not repeating the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'pywt_per'``:  like ``'periodic'``-padding but gives the smallest
            possible number of decomposition coefficients.
            Only available with ``impl='pywt'``, See ``pywt.Modes.modes``.

            ``'reflect'``: Reflect at the boundary, without repeating the
            outmost values.

            ``'antisymmetric'``: Anti-symmetric variant of ``symmetric``.

            ``'antireflect'``: Anti-symmetric variant of ``reflect``.

            For reference, the following table compares the naming conventions
            for the modes in ODL vs. PyWavelets::

                ======================= ==================
                          ODL               PyWavelets
                ======================= ==================
                symmetric               symmetric
                reflect                 reflect
                order1                  smooth
                order0                  constant
                constant, pad_const=0   zero
                periodic                periodic
                pywt_per                periodization
                antisymmetric           antisymmetric
                antireflect             antireflect
                ======================= ==================

            See `signal extension modes`_ for an illustration of the modes
            (under the PyWavelets naming conventions).
        pad_const : float, optional
            Constant value to use if ``pad_mode == 'constant'``. Ignored
            otherwise. Constants other than 0 are not supported by the
            ``pywt`` back-end.
        impl : {'pywt'}, optional
            Back-end for the wavelet transform.
        axes : sequence of ints, optional
            Axes over which the DWT that created ``coeffs`` was performed.  The
            default value of ``None`` corresponds to all axes. When not all
            axes are included this is analagous to a batch transform in
            ``len(axes)`` dimensions looped over the non-transformed axes. In
            orther words, filtering and decimation does not occur along any
            axes not in ``axes``.

        Examples
        --------
        Check that the inverse is the actual inverse on a simple example on
        a discrete 2D space with 4 sampling points per axis:

        >>> space = odl.uniform_discr([0, 0], [1, 1], (4, 4))
        >>> wavelet_trafo = odl.trafos.WaveletTransform(
        ...     domain=space, nlevels=1, wavelet='haar')
        >>> orig_array = space.element(np.array([[1, 1, 1, 1],
        ...                                      [0, 0, 0, 0],
        ...                                      [0, 0, 1, 1],
        ...                                      [1, 0, 1, 0]]))
        >>> decomp = wavelet_trafo(orig_array)
        >>> recon = wavelet_trafo.inverse(decomp)
        >>> from odl.core.util.testutils import all_almost_equal
        >>> all_almost_equal(recon, orig_array)
        True

        References
        ----------
        .. _signal extension modes:
           https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-modes.html
        """
        super(WaveletTransformInverse, self).__init__(
            space=range, wavelet=wavelet, variant='inverse', nlevels=nlevels,
            pad_mode=pad_mode, pad_const=pad_const, impl=impl, axes=axes)

    def _call(self, coeffs):
        """Return the inverse wavelet transform of ``coeffs``."""
        if self.impl == 'pywt':
            coeffs = pywt.unravel_coeffs(coeffs.data,
                                         coeff_slices=self._coeff_slices,
                                         coeff_shapes=self._coeff_shapes,
                                         output_format='wavedecn')
            recon = pywt.waverecn(
                coeffs, wavelet=self.pywt_wavelet, mode=self.pywt_pad_mode,
                axes=self.axes)
            recon_shape = self.range.shape
            if recon.shape != recon_shape:
                # If the original shape was odd along any transformed axes it
                # will have been rounded up to the next even size after the
                # reconstruction. The extra sample should be discarded.
                # The underlying reason is decimation by two in reconstruction
                # must keep ceil(N/2) samples in each band for perfect
                # reconstruction. Reconstruction then upsamples by two.
                # When N is odd, (2 * np.ceil(N/2)) != N.
                recon_slc = []
                for i, (n_recon, n_intended) in enumerate(zip(recon.shape,
                                                              recon_shape)):
                    if n_recon == n_intended + 1:
                        # Upsampling added one entry too much in this axis,
                        # drop last one
                        recon_slc.append(slice(-1))
                    elif n_recon == n_intended:
                        recon_slc.append(slice(None))
                    else:
                        raise ValueError(
                            'in axis {}: expected size {} or {} in '
                            '`recon_shape`, got {}'
                            ''.format(i, n_recon - 1, n_recon,
                                      n_intended))
                recon = recon[tuple(recon_slc)]
            return self.range.element(recon)
        else:
            raise RuntimeError("bad `impl` '{}'".format(self.impl))

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `WaveletTransform`
            If the transform is orthogonal, the adjoint is the inverse.

        Raises
        ------
        OpNotImplementedError
            if `is_orthogonal` is ``False``

        See Also
        --------
        inverse
        """
        if self.is_orthogonal:
            scale = self.range.partition.cell_volume
            return scale * self.inverse
        else:
            # TODO: put adjoint here
            return super(WaveletTransformInverse, self).adjoint

    @property
    def inverse(self):
        """Inverse of this operator.

        Returns
        -------
        inverse : `WaveletTransform`

        See Also
        --------
        adjoint
        """
        return WaveletTransform(
            domain=self.range, wavelet=self.pywt_wavelet, nlevels=self.nlevels,
            pad_mode=self.pad_mode, pad_const=self.pad_const, impl=self.impl,
            axes=self.axes)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests(skip_if=not PYWT_AVAILABLE)
